"""A directory holding files that may or may not be marked for uploading. Updates when new files are added."""

#imports
import pathlib, datetime, time
from threading import Lock
from queue import Queue
from ...kafka_wrapper import ProducerGroup
from ...workflow import Runnable
from ...workflow.controlled_process_single_thread import ControlledProcessSingleThread
from ...utilities.misc import populated_kwargs
from ...utilities.exception_tracking_thread import ExceptionTrackingThread
from ..config import RUN_OPT_CONST
from .. import DataFileDirectory
from .file_registry.producer_file_registry import ProducerFileRegistry
from ..entity.upload_data_file import UploadDataFile

class DataFileUploadDirectory(DataFileDirectory,ControlledProcessSingleThread,ProducerGroup,Runnable) :
    """
    Class representing a directory being watched for new files to be added.
    Files added to the directory are broken into chunks and uploaded as messages to a Kafka topic.

    :param dirpath: Path to the directory that should be watched for new files
    :type dirpath: :class:`pathlib.Path`
    :param config_path: Path to the config file to use in defining the Broker connection and Producers
    :type config_path: :class:`pathlib.Path`
    :param upload_regex: only files matching this regular expression will be uploaded
    :type upload_regex: :func:`re.compile`, optional
    :param datafile_type: the type of data file that recognized files should be uploaded as
        (must be a subclass of :class:`~UploadDataFile`)
    :type datafile_type: :class:`~UploadDataFile`, optional

    :raises ValueError: if `datafile_type` is not a subclass of :class:`~UploadDataFile`
    """

    #################### CONSTANTS ####################

    MIN_WAIT_TIME = 0.005 # starting point for how long to wait between pinging the directory looking for new files
    MAX_WAIT_TIME = 60 # never wait more than a minute to check again for new files
    LOG_SUBDIR_NAME = 'LOGS'

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,dirpath,config_path,upload_regex=RUN_OPT_CONST.DEFAULT_UPLOAD_REGEX,
                    datafile_type=UploadDataFile,**kwargs) :
        """
        Constructor method
        """
        #create a subdirectory for the logs
        self.__logs_subdir = dirpath/self.LOG_SUBDIR_NAME
        if not self.__logs_subdir.is_dir() :
            self.__logs_subdir.mkdir(parents=True)
        #put the log file in the subdirectory
        kwargs = populated_kwargs(kwargs,{'logger_file':self.__logs_subdir})
        super().__init__(dirpath,config_path,**kwargs)
        if not issubclass(datafile_type,UploadDataFile) :
            errmsg = 'ERROR: DataFileUploadDirectory requires a datafile_type that is a subclass of '
            errmsg+= f'UploadDataFile but {datafile_type} was given!'
            self.logger.error(errmsg,ValueError)
        self.__upload_regex = upload_regex
        self.__datafile_type = datafile_type
        self.__wait_time = self.MIN_WAIT_TIME
        self.__lock = Lock()
        self.__topic_name = None
        self.__chunk_size = None
        self.__file_registry = None
        self.__upload_queue = None
        self.__producers = []
        self.__upload_threads = []

    def upload_files_as_added(self,topic_name,
                              n_threads=RUN_OPT_CONST.N_DEFAULT_UPLOAD_THREADS,
                              chunk_size=RUN_OPT_CONST.DEFAULT_CHUNK_SIZE,
                              max_queue_size=RUN_OPT_CONST.DEFAULT_MAX_UPLOAD_QUEUE_SIZE,
                              upload_existing=False) :
        """
        Watch for new files to be added to the directory.
        Chunk and produce newly added files as messages to a Kafka topic.

        :param topic_name: Name of the topic to produce messages to
        :type topic_name: str
        :param n_threads: The number of threads to use to produce from the shared queue
        :type n_threads: int, optional
        :param chunk_size: The size of each file chunk in bytes
        :type chunk_size: int, optional
        :param max_queue_size: The maximum number of items allowed to be placed in the internal upload queue at once
        :type max_queue_size: int, optional
        :param upload_existing: True if any files that already exist in the directory should be uploaded. If False
            (the default) then only files added to the directory after startup will be enqueued to the producer
        :type upload_existing: bool, optional
        """
        #set the topic name and chunk size
        self.__topic_name = topic_name
        self.__chunk_size = chunk_size
        #start up a file registry in the watched directory
        self.__file_registry = ProducerFileRegistry(dirpath=self.__logs_subdir,topic_name=topic_name,logger=self.logger)
        #if we're only going to upload new files, exclude what's already in the directory
        if not upload_existing :
            self.__find_new_files(to_upload=False)
        #add or modify datafiles to upload based on the contents of the file registry
        self.__startup_from_file_registry()
        #start the upload queue and thread
        msg = 'Will upload '
        if not upload_existing :
            msg+='new files added to '
        else :
            msg+='files in '
        msg+=f'{self.dirpath} to the {topic_name} topic using {n_threads} threads'
        self.logger.info(msg)
        self.__upload_queue = Queue(max_queue_size)
        #start the producers and upload threads
        for _ in range(n_threads) :
            self.__producers.append(self.get_new_producer())
            thread = ExceptionTrackingThread(target=self.__producers[-1].produce_from_queue_looped,
                                             args=(self.__upload_queue,self.__topic_name),
                                             kwargs={'callback': self.producer_callback},
                    )
            thread.start()
            self.__upload_threads.append(thread)
        #loop until the user inputs a command to stop
        self.run()
        #return a list of filepaths that have been uploaded
        return [fp for fp,datafile in self.data_files_by_path.items() if datafile.fully_produced]

    def producer_callback(self,err,msg,filename,filepath,n_total_chunks,chunk_i) :
        """
        A reference to this method is given as the callback for each call to :func:`confluent_kafka.Producer.produce`.
        It is called for every message upon acknowledgement by the broker, and it uses the file registries in the
        LOGS subdirectory to keep the information about what has and hasn't been uploaded current with what has
        been received by the broker.

        If the message is the final one to be acknowledged from its corresponding :class:`~UploadDataFile`,
        the :class:`~UploadDataFile` is registered as "fully produced".

        :param err: The error object for the message
        :type err: :class:`confluent_kafka.KafkaError`
        :param msg: The message object
        :type msg: :class:`confluent_kafka.Message`
        :param filename: The name of the file the message is coming from
        :type filename: str
        :param filepath: The full path to the file the message is coming from
        :type filepath: :class:`pathlib.Path`
        :param n_total_chunks: The total number of chunks in the file this message came from
        :type n_total_chunks: int
        :param chunk_i: The index of the message's file chunk in the full list for the file
        :type chunk_i: int
        """
        # If any error occured, re-enqueue the message's file chunk
        if err is None and msg.error() is not None :
            err = msg.error()
        if err is not None :
            if err.fatal() :
                warnmsg =f'WARNING: fatally failed to deliver message for chunk {chunk_i} of {filepath}. '
                warnmsg+=f'This message will be re-enqeued. Error reason: {err.str()}'
            elif not err.retriable() :
                warnmsg =f'WARNING: Failed to deliver message for chunk {chunk_i} of {filepath} and cannot retry. '
                warnmsg+= f'This message will be re-enqueued. Error reason: {err.str()}'
            self.logger.warning(warnmsg)
            self.__add_chunks_for_filepath(filepath,[chunk_i])
        # Otherwise, register the chunk as successfully sent to the broker
        else :
            with self.__lock :
                fully_produced = self.__file_registry.register_chunk(filename,filepath,n_total_chunks,chunk_i)
                #If the file has now been fully produced to the topic, set the variable for the file and log a line
                if fully_produced :
                    self.data_files_by_path[filepath].fully_produced = True
                    infomsg = f'{filepath.relative_to(self.dirpath)} has been fully produced to the '
                    infomsg+= f'"{self.__topic_name}" topic as '
                    if n_total_chunks==1 :
                        infomsg+=f'{n_total_chunks} message'
                    else :
                        infomsg+=f'a set of {n_total_chunks} messages'
                    self.logger.info(infomsg)

    def filepath_should_be_uploaded(self,filepath) :
        """
        Returns True if a given filepath should be uploaded (if it matches the upload regex)

        :param filepath: A candidate upload filepath
        :type filepath: :class:`pathlib.Path`

        :return: True if the filepath is an existing file outside of the LOGS directory
            whose name matches the upload regex, False otherwise
        :rtype: bool

        :raises TypeError: if `filepath` isn't a :class:`pathlib.Path` object
        """
        if not isinstance(filepath,pathlib.PurePath) :
            self.logger.error(f'ERROR: {filepath} passed to filepath_should_be_uploaded is not a Path!',TypeError)
        if not filepath.is_file() :
            return False
        if filepath.parent == self.__logs_subdir :
            return False
        if self.__upload_regex.match(filepath.name) :
            return True
        return False

    #################### PRIVATE HELPER FUNCTIONS ####################

    def _run_iteration(self) :
        #check for new files in the directory if we haven't already found some to run
        if not self.have_file_to_upload :
            # wait here with some slight backoff so that the watched directory isn't just constantly pinged
            time.sleep(self.__wait_time)
            self.__find_new_files()
            n_new_callbacks = 0
            for producer in self.__producers :
                new_callbacks = producer.poll(0)
                if new_callbacks is not None :
                    n_new_callbacks+=new_callbacks
                if n_new_callbacks>0 :
                    break
            if (not self.have_file_to_upload) and (n_new_callbacks==0) :
                if self.__wait_time<self.MAX_WAIT_TIME :
                    self.__wait_time*=1.1
            else :
                self.__wait_time = self.MIN_WAIT_TIME
            return
        #find the first file that's running and add some of its chunks to the upload queue
        for datafile in self.data_files_by_path.values() :
            if datafile.upload_in_progress or datafile.waiting_to_upload :
                datafile.enqueue_chunks_for_upload(self.__upload_queue,
                                                    n_threads=len(self.__upload_threads),
                                                    chunk_size=self.__chunk_size)
        #check for new files again
        self.__find_new_files()
        #restart any crashed threads
        self.__restart_crashed_threads()

    def _on_check(self) :
        #poll the producers
        for producer in self.__producers :
            producer.poll(0)
        #log progress so far
        self.logger.debug(self.progress_msg)
        #reset the wait time
        self.__wait_time = self.MIN_WAIT_TIME

    def _on_shutdown(self) :
        self.logger.info('Will quit after all currently enqueued files are done being transferred.')
        self.logger.info(self.progress_msg)
        #add the remainder of any files currently in progress
        if self.n_partially_done_files>0 :
            msg='Will finish queueing the remainder of the following files before flushing the producer and quitting:\n'
            for pdfp in self.partially_done_file_paths :
                msg+=f'\t{pdfp}\n'
            self.logger.info(msg)
        while self.n_partially_done_files>0 :
            for datafile in self.data_files_by_path.values() :
                if datafile.upload_in_progress :
                    datafile.enqueue_chunks_for_upload(self.__upload_queue,
                                                        n_threads=len(self.__upload_threads),
                                                        chunk_size=self.__chunk_size)
                    break
        #stop the uploading threads by adding "None"s to their queues and joining them
        for thread in self.__upload_threads :
            self.__upload_queue.put(None)
        for thread in self.__upload_threads :
            thread.join()
        self.logger.info('Waiting for all enqueued messages to be delivered (this may take a moment)....')
        for producer in self.__producers :
            producer.flush(timeout=-1) #don't move on until all enqueued messages have been sent/received
            producer.close()
        self.close()

    def __find_new_files(self,to_upload=True) :
        """
        Search the directory for any unrecognized files and add them to _data_files_by_path

        to_upload = if False, new files found will NOT be marked for uploading
                    (default is new files are expected to be uploaded)
        """
        #This is in a try/except in case a file is moved or a subdirectory is renamed while this method is running
        #it'll just return and try again if so
        try :
            for filepath in self.dirpath.rglob('*') :
                filepath = filepath.resolve()
                if (filepath not in self.data_files_by_path) and self.filepath_should_be_uploaded(filepath) :
                    #wait until the file is actually available
                    file_ready = False
                    while not file_ready :
                        try :
                            with open(filepath) as _ :
                                file_ready = True
                        except PermissionError :
                            time.sleep(0.05)
                    self.data_files_by_path[filepath]=self.__datafile_type(filepath,
                                                                          to_upload=to_upload,
                                                                          rootdir=self.dirpath,
                                                                          logger=self.logger,
                                                                          **self.other_datafile_kwargs)
        except FileNotFoundError :
            return

    def __startup_from_file_registry(self) :
        """
        Look through the file registry to find and prepare to enqueue any files
        that were in progress the last time the code was shut down and ignore
        any files that have previously been fuly uploaded
        """
        #make sure any files listed as "in progress" have their remaining chunks enqueued
        for filepath,chunks in self.__file_registry.get_incomplete_filepaths_and_chunks() :
            self.logger.info(f'Found {filepath} in progress from a previous run. Will re-enqueue {len(chunks)} chunks.')
            self.__add_chunks_for_filepath(filepath,chunks)
        #make sure any files listed as "completed" will not be uploaded again
        for filepath in self.__file_registry.get_completed_filepaths() :
            if filepath in self.data_files_by_path :
                if self.data_files_by_path[filepath].to_upload :
                    msg = f'Found {filepath} listed as fully uploaded during a previous run. Will not produce it again.'
                    self.logger.info(msg)
                self.data_files_by_path[filepath].to_upload=False
            elif filepath.is_file() :
                msg = f'Found {filepath} listed as fully uploaded during a previous run. Will not produce it again.'
                self.logger.info(msg)
                self.data_files_by_path[filepath]=self.__datafile_type(filepath,
                                                                       to_upload=False,
                                                                       rootdir=self.dirpath,
                                                                       logger=self.logger,
                                                                       **self.other_datafile_kwargs)

    def __add_chunks_for_filepath(self,filepath,chunks) :
        """
        Add the given chunks to the list of chunks to upload for a given file
        Creates the file if it doesn't already exist in the dictionary of recognized files
        """
        if filepath in self.data_files_by_path :
            self.data_files_by_path[filepath].to_upload=True
        else :
            self.data_files_by_path[filepath]=self.__datafile_type(filepath,
                                                                    to_upload=True,
                                                                    rootdir=self.dirpath,
                                                                    logger=self.logger,
                                                                    **self.other_datafile_kwargs)
        self.data_files_by_path[filepath].add_chunks_to_upload(chunks,chunk_size=self.__chunk_size)

    def __restart_crashed_threads(self) :
        """
        Iterate over the upload threads and restart any that crashed, logging the exception they threw
        """
        for ti,upload_thread in enumerate(self.__upload_threads) :
            if upload_thread.caught_exception is not None :
                #log the error
                warnmsg = 'WARNING: an upload thread raised an Exception, which will be logged as an error below '
                warnmsg = 'but not reraised. The Producer and upload thread that raised the error will be restarted.'
                self.logger.warning(warnmsg)
                self.logger.log_exception_as_error(upload_thread.caught_exception,reraise=False)
                #try to join the thread
                try :
                    upload_thread.join()
                except Exception :
                    pass
                finally :
                    self.__upload_threads[ti] = None
                    self.__producers[ti] = None
                #recreate the producer and restart the thread
                self.__producers[ti] = self.get_new_producer()
                thread = ExceptionTrackingThread(target=self.__producers[ti].produce_from_queue_looped,
                                            args=(self.__upload_queue,self.__topic_name),
                                            kwargs={'callback': self.producer_callback},
                            )
                thread.start()
                self.__upload_threads[ti] = thread

    #################### CLASS METHODS ####################

    @classmethod
    def get_command_line_arguments(cls) :
        superargs,superkwargs = super().get_command_line_arguments()
        args = [*superargs,'upload_dir','config','topic_name','upload_regex',
                'chunk_size','queue_max_size','update_seconds','upload_existing']
        kwargs = {**superkwargs,'n_threads':RUN_OPT_CONST.N_DEFAULT_UPLOAD_THREADS}
        return args, kwargs

    @classmethod
    def run_from_command_line(cls,args=None) :
        """
        Run a :class:`~DataFileUploadDirectory` directly from the command line

        Calls :func:`~upload_files_as_added` on a :class:`~DataFileUploadDirectory` defined by
        command line (or given) arguments

        :param args: the list of arguments to send to the parser instead of getting them from sys.argv
        :type args: List
        """
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        #make the DataFileDirectory for the specified directory
        upload_file_directory = cls(args.upload_dir,args.config,
                                    upload_regex=args.upload_regex,update_secs=args.update_seconds)
        #listen for new files in the directory and run uploads as they come in until the process is shut down
        run_start = datetime.datetime.now()
        if not args.upload_existing :
            upload_file_directory.logger.info(f'Listening for files to be added to {args.upload_dir}...')
        else :
            upload_file_directory.logger.info(f'Uploading files in/added to {args.upload_dir}...')
        uploaded_filepaths = upload_file_directory.upload_files_as_added(args.topic_name,
                                                                         n_threads=args.n_threads,
                                                                         chunk_size=args.chunk_size,
                                                                         max_queue_size=args.queue_max_size,
                                                                         upload_existing=args.upload_existing)
        run_stop = datetime.datetime.now()
        upload_file_directory.logger.info(f'Done listening to {args.upload_dir} for files to upload')
        final_msg = f'The following {len(uploaded_filepaths)} file'
        if len(uploaded_filepaths)==1 :
            final_msg+=' was'
        else :
            final_msg+='s were'
        final_msg+=f' uploaded between {run_start} and {run_stop}:\n'
        for fp in uploaded_filepaths :
            final_msg+=f'\t{fp.relative_to(args.upload_dir)}\n'
        upload_file_directory.logger.info(final_msg)

    #################### PROPERTIES ####################

    @property
    def other_datafile_kwargs(self) :
        """
        Overload this property to send extra keyword arguments to the :class:`~UploadDataFile` constructor
        for each recognized file (useful if using a custom `datafile_type`)
        """
        return {}
    @property
    def progress_msg(self) :
        """
        A message describing the files that are currently recognized as part of the directory
        """
        self.__find_new_files()
        progress_msg = 'The following files have been recognized so far:\n'
        for datafile in self.data_files_by_path.values() :
            if not datafile.to_upload :
                continue
            progress_msg+=f'\t{datafile.upload_status_msg}\n'
        return progress_msg
    @property
    def have_file_to_upload(self) :
        """
        True if there are any files waiting to be uploaded or in progress
        """
        for datafile in self.data_files_by_path.values() :
            if datafile.upload_in_progress or datafile.waiting_to_upload :
                return True
        return False
    @property
    def partially_done_file_paths(self) :
        """
        A list of filepaths whose uploads are currently in progress
        """
        return [fp for fp,df in self.data_files_by_path.items() if df.upload_in_progress]
    @property
    def n_partially_done_files(self) :
        """
        The number of files currently in the process of being uploaded
        """
        return len(self.partially_done_file_paths)

def main(args=None) :
    """
    Main method to run from command line
    """
    DataFileUploadDirectory.run_from_command_line(args)

if __name__=='__main__' :
    main()



"""Example of a general script that can be installed as a Service/daemon"""

#imports
import pathlib, datetime, time
from argparse import ArgumentParser

def main(args=None) :
    """
    The main function that will be run as a Service/daemon
    """
    parser = ArgumentParser()
    parser.add_argument('output_dir',type=pathlib.Path,
                        help='Path to the directory where the test file should be created')
    args = parser.parse_args(args)
    if not args.output_dir.is_dir() :
        if args.output_dir.exists() :
            raise ValueError(f'ERROR: given output directory {args.output_dir} exists but is not a directory!')
        args.output_dir.mkdir(parents=True)
    test_file_name = 'script_example_service_test.txt'
    if not (args.output_dir/test_file_name).is_file() :
        with open(args.output_dir/test_file_name,'w') as fp :
            fp.write('This file was created to test running a generic Python script as a Service/daemon')
    with open(args.output_dir/test_file_name,'a') as fp :
        fp.write(f'\nService rerun {(datetime.datetime.now()).strftime("on %Y-%m-%d at %H:%M:%S")}')
    time.sleep(30)

if __name__=='__main__' :
    main()



"""Various helper functions for working with Services/daemons"""

#imports
import pathlib, os, sys, platform
from subprocess import check_output, CalledProcessError
from .config import SERVICE_CONST

def get_os_name() :
    """
    Return the name of the operating system the Service is being installed or running on
    """
    if platform.system()=='Windows' :
        return 'Windows'
    if platform.system()=='Linux' :
        return 'Linux'
    #MacOS is not supported
    if platform.system()=='Darwin' :
        errmsg = 'ERROR: Installing programs as Services is not supported on MacOS!'
        SERVICE_CONST.logger.error(errmsg,NotImplementedError)
    #otherwise I don't know what happened
    else :
        errmsg = f'ERROR: could not determine operating system from platform.system() output "{platform.system()}"'
        SERVICE_CONST.logger.error(errmsg,ValueError)
    return None

def run_cmd_in_subprocess(args,*,shell=False,logger=None) :
    """
    run a command in a subprocess and return its result, printing and re-throwing any exceptions it causes
    """
    if isinstance(args,str) :
        args = [args]
    try :
        result = check_output(args,shell=shell,env=os.environ)
        return result
    except CalledProcessError as exc :
        errmsg = 'ERROR: failed to run a command! '
        if exc.output is not None and exc.output.strip()!='' :
            errmsg+= f'\noutput:\n{exc.output.decode()}'
        if exc.stdout is not None and exc.stdout.strip()!=''  :
            errmsg+= f'\nstdout:\n{exc.stdout.decode()}'
        if exc.stderr is not None and exc.stderr.strip()!='' :
            errmsg+= f'\nstderr:\n{exc.stderr.decode()}'
        if logger is not None :
            logger.error(errmsg,exc_obj=exc)
        else :
            SERVICE_CONST.logger.error(errmsg,exc_obj=exc)
        return None

def set_env_var(var_name,var_val) :
    """
    set an environment variable given its name and value
    """
    if get_os_name()=='Windows' :
        pwrsh_cmd = f'[Environment]::SetEnvironmentVariable("{var_name}","{var_val}"'
        pwrsh_cmd+= ',[EnvironmentVariableTarget]::Machine)'
        run_cmd_in_subprocess(['powershell.exe',pwrsh_cmd])
    elif get_os_name()=='Linux' :
        run_cmd_in_subprocess(['export',f'{var_name}={var_val}'],shell=True)
    os.environ[var_name]=var_val

def remove_env_var(var_name) :
    """
    remove an environment variable given its name
    """
    if get_os_name()=='Windows' :
        pwrsh_cmd = f'[Environment]::SetEnvironmentVariable("{var_name}",$null,[EnvironmentVariableTarget]::Machine)'
        run_cmd_in_subprocess(['powershell.exe',pwrsh_cmd])
    elif get_os_name()=='Linux' :
        run_cmd_in_subprocess(['unset',var_name],shell=True)
    else :
        raise NotImplementedError
    os.environ[var_name]=''

def set_env_var_from_user_input(var_name) :
    """
    set an environment variable with the given name and description based on user input
    """
    var_val = input(f'Please enter the value for environment variable {var_name}: ')
    set_env_var(var_name,var_val)

def set_env_vars(env_var_names,interactive=True) :
    """
    set the necessary environment variables
    """
    variables_set = False
    for env_var_name in env_var_names :
        if (not interactive) and (env_var_name in ('KAFKA_PROD_CLUSTER_USERNAME','KAFKA_PROD_CLUSTER_PASSWORD')) :
            continue
        if os.path.expandvars(f'${env_var_name}') == f'${env_var_name}' :
            if interactive :
                set_env_var_from_user_input(env_var_name)
                variables_set = True
            else :
                raise RuntimeError(f'ERROR: a value for the {env_var_name} environment variable is not set!')
        else :
            if interactive :
                choice = input(f'A value for the {env_var_name} is already set, would you like to reset it? [y/(n)]: ')
                if choice.lower() in ('yes','y') :
                    set_env_var_from_user_input(env_var_name)
                    variables_set = True
    return (variables_set and get_os_name()=='Windows')

def test_python_code() :
    """
    briefly test the python code for the repo to catch any errors
    """
    must_rerun = set_env_vars(['KAFKA_TEST_CLUSTER_USERNAME','KAFKA_TEST_CLUSTER_PASSWORD'])
    if must_rerun :
        msg = 'New values for environment variables have been set. '
        msg+= 'Please close this window and rerun InstallService so that their values get picked up.'
        SERVICE_CONST.logger.info(msg)
        sys.exit(0)
    SERVICE_CONST.logger.debug('Testing code to check for errors...')
    unittest_dir_path = pathlib.Path(__file__).parent.parent.parent / 'test' / 'unittests'
    SERVICE_CONST.logger.debug(f'Running all unittests in {unittest_dir_path}...')
    run_cmd_in_subprocess([f'{sys.executable}','-m','unittest','discover','-s',f'{unittest_dir_path}','-vf'])
    SERVICE_CONST.logger.debug('All unittest checks complete : )')



"""Code for installing OpenMSIStream programs (or others) as Windows Services/Linux daemons"""

from .windows_service_manager import WindowsServiceManager
from .linux_service_manager import LinuxServiceManager

__all__ = [
    'WindowsServiceManager',
    'LinuxServiceManager',
]



"""Wrapper around the KafkaCrypto producer/consumer pair communicating with the key passing topics"""

#imports
import pathlib, uuid, warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from kafkacrypto import KafkaProducer, KafkaConsumer, KafkaCrypto
from ..utilities.misc import change_dir

class OpenMSIStreamKafkaCrypto :
    """
    A wrapper class to own and work with the Producers, Consumers, and other objects needed
    by KafkaCrypto to handle encrypting/decrypting messages and exchanging keys

    :param broker_configs: the producer/consumer-agnostic configurations for connecting to the broker
    :type broker_configs: dict
    :param config_file: the path to the KafkaCrypto config file for the node being used
    :type config_file: :class:`pathlib.Path`
    """

    @property
    def config_file_path(self) :
        """
        The path to the KafkaCrypto config file used
        """
        return self.__config_file
    @property
    def key_serializer(self) :
        """
        The crypto key serializer
        """
        return self._kc.getKeySerializer()
    @property
    def key_deserializer(self) :
        """
        The crypto key deserializer
        """
        return self._kc.getKeyDeserializer()
    @property
    def value_serializer(self) :
        """
        The crypto value serializer
        """
        return self._kc.getValueSerializer()
    @property
    def value_deserializer(self) :
        """
        The crypto value deserializer
        """
        return self._kc.getValueDeserializer()

    def __init__(self,broker_configs,config_file) :
        """
        Constructor method
        """
        producer_configs = broker_configs.copy()
        consumer_configs = broker_configs.copy()
        #figure out a consumer group ID to use (KafkaCrypto Consumers need one)
        if 'group.id' not in consumer_configs.keys() :
            consumer_configs['group.id'] = str(uuid.uuid1())
        with change_dir(pathlib.Path(config_file).parent) :
            #start up the producer and consumer
            self._kcp = KafkaProducer(**producer_configs)
            self._kcc = KafkaConsumer(**consumer_configs)
            #initialize the KafkaCrypto object
            self._kc = KafkaCrypto(None,self._kcp,self._kcc,config_file)
        self.__config_file = config_file

    def close(self) :
        """
        Close the :class:`kafkacrypto.KafkaCrypto` object and un-set its producer/consumer
        """
        try :
            self._kc.close()
        except Exception :
            pass
        finally :
            self._kc = None
            self._kcp = None
            self._kcc = None



"""Various utility functions used elsewhere in this subdirectory"""

#imports
import datetime
from confluent_kafka import TIMESTAMP_NOT_AVAILABLE, TIMESTAMP_CREATE_TIME, TIMESTAMP_LOG_APPEND_TIME

def get_encrypted_message_timestamp_string(msg) :
    """
    Given a Message object containing still-encrypted keys and/or values,
    return a string describing the timestamp in local time.
    """
    format_string = "on_%b_%d_%Y_at_%H_%M_%S_%f"
    ts_type, timestamp = msg.timestamp
    if ts_type==TIMESTAMP_NOT_AVAILABLE :
        timestamp_string = f'consumed_{datetime.datetime.now().strftime(format_string)}'
    elif ts_type in (TIMESTAMP_CREATE_TIME,TIMESTAMP_LOG_APPEND_TIME) :
        utc_dt = datetime.datetime.fromtimestamp(timestamp/1000.,tz=datetime.timezone.utc)
        local_timezone = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
        local_dt = utc_dt.astimezone(local_timezone)
        if ts_type==TIMESTAMP_CREATE_TIME :
            timestamp_string = 'produced'
        elif ts_type==TIMESTAMP_LOG_APPEND_TIME :
            timestamp_string = 'received'
        timestamp_string+=f'_{local_dt.strftime(format_string)}'
    else :
        raise ValueError(f'ERROR: Unrecognized timestamp type {ts_type}')
    return timestamp_string

def get_encrypted_message_key_and_value_filenames(msg,topic_name) :
    """
    Given a Message object containing still-encrypted keys and/or values and the name
    of the topic it came from, return the names of the files to which the key and
    value should have their bytes written
    """
    timestamp_string = get_encrypted_message_timestamp_string(msg)
    fp_pp = f'encrypted_message_from_{topic_name}_{timestamp_string}'
    return f'{fp_pp}_key.bin',f'{fp_pp}_value.bin'

def get_message_prepend(subdir_str,filename) :
    """
    Return the prepend for each message key, based on the filepath
    """
    key_pp = ''
    if subdir_str is not None :
        key_pp = f'{"_".join(subdir_str.split("/"))}'
    if key_pp!='' :
        key_pp+='_'
    return f'{key_pp}{filename}_chunk'



"""Manage Windows Services"""

#imports
import sys, os, pathlib, shutil, ctypes.util
from subprocess import CalledProcessError
from .config import SERVICE_CONST
from .utilities import run_cmd_in_subprocess
from .service_manager_base import ServiceManagerBase

class WindowsServiceManager(ServiceManagerBase) :
    """
    Base class for working with Windows Services

    :param service_name: The name of the Service as installed
    :type service_name: str
    :param service_spec_string: A string specifying which code should be run as a Service.
        Could be the name of an OpenMSIStream Runnable class, or the path to a custom Python code.
        Custom Services can also specify a :class:`~.workflow.Runnable` class name,
        and/or a function in the file using special formatting like [class_name]=[path.to.file]:[function_name].
        Only needed to initially install the Service.
    :type service_spec_string: str, optional
    :param argslist: The list of arguments (as from the command line) to pass to the
        :class:`~.workflow.Runnable` class.
        Only needed to initially install the Service.
    :type argslist: list, optional
    :param interactive: if True, a few more messages/prompts will come up telling a user what to do
    :type interactive: bool, optional
    """

    @property
    def env_var_names(self):
        for evn in super().env_var_names :
            yield evn
        #get the names of environment variables from the env_var file
        if self.env_var_filepath.is_file() :
            with open(self.env_var_filepath,'r') as fp :
                lines = fp.readlines()
            for line in lines :
                linestrip = line.strip()
                if linestrip!='' :
                    yield linestrip

    def install_service(self) :
        """
        Install the Service
        """
        super().install_service()
        #if it doesn't exist there yet, copy the libsodium.dll file to C:\Windows\system32
        self.__copy_lib_dlls_to_system32()
        #find or install NSSM to run the executable
        self.__find_install_nssm()
        #run NSSM to install the service
        cmd = f'{SERVICE_CONST.NSSM_PATH} install {self.service_name} \"{sys.executable}\" \"{self.exec_fp}\"'
        run_cmd_in_subprocess(['powershell.exe',cmd],logger=self.logger)
        run_cmd_in_subprocess(['powershell.exe',
                               f'{SERVICE_CONST.NSSM_PATH} set {self.service_name} DisplayName {self.service_name}'],
                               logger=self.logger)
        self.logger.info(f'Done installing {self.service_name}')

    def start_service(self) :
        """
        Start the Service
        """
        self.logger.info(f'Starting {self.service_name}...')
        #start the service using net
        cmd = f'net start {self.service_name}'
        run_cmd_in_subprocess(['powershell.exe',cmd],logger=self.logger)
        self.logger.info(f'Done starting {self.service_name}')

    def service_status(self) :
        """
        Print the status of the Service
        """
        #find or install NSSM in the current directory
        self.__find_install_nssm()
        #get the service status
        cmd = f'{SERVICE_CONST.NSSM_PATH} status {self.service_name}'
        result = run_cmd_in_subprocess(['powershell.exe',cmd],logger=self.logger)
        self.logger.info(f'{self.service_name} status: {result.decode()}')

    def stop_service(self) :
        """
        Stop the Service
        """
        self.logger.info(f'Stopping {self.service_name}...')
        #stop the service using net
        cmd = f'net stop {self.service_name}'
        run_cmd_in_subprocess(['powershell.exe',cmd],logger=self.logger)
        self.logger.info(f'Done stopping {self.service_name}')

    def remove_service(self,remove_env_vars=False,remove_install_args=False,remove_nssm=False) :
        """
        Remove the Service

        :param remove_env_vars: if True, any environment variables needed by the Service will be removed.
        :type remove_env_vars: bool, optional.
        :param remove_install_args: if True, the file listing the arguments used to install the Service
            (to make it easier to re-install) will be removed.
        :type remove_install_args: bool, optional
        :param remove_nssm: if True, the NSSM executable will be removed.
        :type remove_nssm: bool, optional
        """
        self.logger.info(f'Removing {self.service_name}...')
        #find or install NSSM in the current directory
        self.__find_install_nssm()
        #using NSSM
        cmd = f'{SERVICE_CONST.NSSM_PATH} remove {self.service_name} confirm'
        run_cmd_in_subprocess(['powershell.exe',cmd],logger=self.logger)
        self.logger.info('Service removed')
        #remove NSSM if requested
        if remove_nssm :
            if SERVICE_CONST.NSSM_PATH.is_file() :
                try :
                    run_cmd_in_subprocess(['powershell.exe',f'del {SERVICE_CONST.NSSM_PATH}'],logger=self.logger)
                except CalledProcessError :
                    msg = f'WARNING: failed to delete {SERVICE_CONST.NSSM_PATH}. '
                    msg+= 'You are free to delete it manually if you would like.'
                    self.logger.info(msg)
            else :
                msg = f'NSSM does not exist at {SERVICE_CONST.NSSM_PATH} so it will not be removed'
                self.logger.info(msg)
        else :
            if SERVICE_CONST.NSSM_PATH.is_file() :
                msg = f'NSSM executable at {SERVICE_CONST.NSSM_PATH} will be retained'
                self.logger.info(msg)
        super().remove_service(remove_env_vars=remove_env_vars,remove_install_args=remove_install_args)

    def _write_env_var_file(self) :
        code = ''
        for evn in self.env_var_names :
            val = os.path.expandvars(f'${evn}')
            if val==f'${evn}' :
                raise RuntimeError(f'ERROR: value not found for expected environment variable {evn}!')
            code += f'{evn}\n'
        if code=='' :
            return False
        with open(self.env_var_filepath,'w') as fp :
            fp.write(code)
        return True

    def __copy_lib_dlls_to_system32(self) :
        """
        Ensure that several .dll files exist in C:\\Windows\\system32
        (Needed to properly load it when running as a service)
        """
        system32_path = pathlib.Path(r'C:\Windows\system32')
        package_names = ['libsodium','libcrypto-1_1-x64','libssl-1_1-x64']
        for pname in package_names :
            current_env_dll = ctypes.util.find_library(pname)
            if current_env_dll is not None :
                current_env_dll_path = pathlib.Path(current_env_dll)
                if (system32_path/current_env_dll_path.name).is_file() :
                    continue
                try :
                    shutil.copy(current_env_dll_path,system32_path/current_env_dll_path.name)
                except Exception as exc :
                    errmsg = f'ERROR: failed to copy {pname} DLL file from {current_env_dll_path} to '
                    errmsg+= f'{system32_path}. This will likely cause the Python code running as a Service '
                    errmsg+=  'to crash. Exception will be logged below, but not reraised.'
                    self.logger.error(errmsg,exc_obj=exc,reraise=False)
            else :
                warnmsg = f'WARNING: could not locate {pname} DLL file to copy to system32 folder for '
                warnmsg+= f'{self.service_name}! This will likely cause the Python code running as a Service to crash.'
                self.logger.warning(warnmsg)

    def __find_install_nssm(self,move_local=True) :
        """
        Ensure the NSSM executable exists in the expected location.
        If it exists in the current directory, move it to the expected location.
        If it doesn't exist anywhere, try to download it from the web, but that's finicky in powershell.
        """
        if SERVICE_CONST.NSSM_PATH.is_file() :
            return
        if (pathlib.Path()/SERVICE_CONST.NSSM_PATH.name).is_file() and move_local :
            (pathlib.Path()/SERVICE_CONST.NSSM_PATH.name).replace(SERVICE_CONST.NSSM_PATH)
            self.__find_install_nssm(move_local=False)
        SERVICE_CONST.logger.info(f'Installing NSSM from {SERVICE_CONST.NSSM_DOWNLOAD_URL}...')
        nssm_zip_file_name = SERVICE_CONST.NSSM_DOWNLOAD_URL.rsplit('/', maxsplit=1)[-1]
        run_cmd_in_subprocess(['powershell.exe',
                                '[Net.ServicePointManager]::SecurityProtocol = [Net.SecurityProtocolType]::Tls12'],
                                logger=self.logger)
        cmd_tuples = [
            (f'curl {SERVICE_CONST.NSSM_DOWNLOAD_URL} -O',
            f'Invoke-WebRequest -Uri {SERVICE_CONST.NSSM_DOWNLOAD_URL} -OutFile {nssm_zip_file_name}'),
            (f'tar -xf {pathlib.Path() / nssm_zip_file_name}',
            f'Expand-Archive {nssm_zip_file_name} -DestinationPath {pathlib.Path().resolve()}'),
            (f'del {nssm_zip_file_name}',
            f'Remove-Item -Path {nssm_zip_file_name}'),
            (f'move {pathlib.Path() / nssm_zip_file_name.rstrip(".zip") / "win64" / "nssm.exe"} {pathlib.Path()}',
            f'''Move-Item -Path {pathlib.Path()/nssm_zip_file_name.rstrip(".zip")/"win64"/"nssm.exe"} \
                -Destination {SERVICE_CONST.NSSM_PATH}'''),
            (f'rmdir /S /Q {nssm_zip_file_name.rstrip(".zip")}',
            f'Remove-Item -Recurse -Force {nssm_zip_file_name.rstrip(".zip")}'),
        ]
        for cmd in cmd_tuples :
            try :
                run_cmd_in_subprocess(['powershell.exe',cmd[1]],logger=self.logger)
            except CalledProcessError :
                run_cmd_in_subprocess(cmd[0],shell=True,logger=self.logger)
        SERVICE_CONST.logger.debug('Done installing NSSM')



"""
Write DataFileChunks directly to disk as they are consumed from topics.
Preserve subdirectory structure if applicable
"""

#imports
import datetime, warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from kafkacrypto.message import KafkaCryptoMessage
from ...workflow import Runnable
from ..config import DATA_FILE_HANDLING_CONST, RUN_OPT_CONST
from ..utilities import get_encrypted_message_key_and_value_filenames
from .. import DataFileDirectory, DownloadDataFileToDisk
from .data_file_chunk_handlers import DataFileChunkProcessor

class DataFileDownloadDirectory(DataFileDirectory,DataFileChunkProcessor,Runnable) :
    """
    Class representing a directory into which files are being reconstructed.

    :param dirpath: Path to the directory where reconstructed files should be saved
    :type dirpath: :class:`pathlib.Path`
    :param config_path: Path to the config file to use in defining the Broker connection and Consumers
    :type config_path: :class:`pathlib.Path`
    :param topic_name: Name of the topic to which the Consumers should be subscribed
    :type topic_name: str
    :param datafile_type: the type of data file that recognized files should be reconstructed as
        (must be a subclass of :class:`~.data_file_io.DownloadDataFileToDisk`)
    :type datafile_type: :class:`~.data_file_io.DownloadDataFileToDisk`, optional

    :raises ValueError: if `datafile_type` is not a subclass of
        :class:`~.data_file_io.DownloadDataFileToDisk`
    """

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,dirpath,config_path,topic_name,datafile_type=DownloadDataFileToDisk,**kwargs) :
        """
        datafile_type = the type of datafile that the consumed messages should be assumed to represent
        In this class datafile_type should be something that extends DownloadDataFileToDisk
        """
        super().__init__(dirpath,config_path,topic_name,datafile_type=datafile_type,**kwargs)
        if not issubclass(self.datafile_type,DownloadDataFileToDisk) :
            errmsg = 'ERROR: DataFileDownloadDirectory requires a datafile_type that is a subclass of '
            errmsg+= f'DownloadDataFileToDisk but {self.datafile_type} was given!'
            self.logger.error(errmsg,ValueError)
        self.__encrypted_messages_subdir = self.dirpath/'ENCRYPTED_MESSAGES'

    def reconstruct(self) :
        """
        Consumes messages and writes their data to disk using several parallel threads to reconstruct the files
        to which they correspond. Runs until the user inputs a command to shut it down.

        :return: the total number of messages consumed
        :rtype: int
        :return: the total number of message processed (written to disk)
        :rtype: int
        :return: the number of files whose reconstruction was completed during the run
        :rtype: int
        """
        msg = f'Will reconstruct files from messages in the {self.topic_name} topic using {self.n_threads} '
        msg+= f'thread{"s" if self.n_threads!=1 else ""}'
        self.logger.info(msg)
        self.run()
        return self.n_msgs_read, self.n_msgs_processed, self.completely_processed_filepaths

    #################### PRIVATE HELPER FUNCTIONS ####################

    def _process_message(self, lock, msg, rootdir_to_set=None):
        retval = super()._process_message(lock,msg,self.dirpath if rootdir_to_set is None else rootdir_to_set)
        #if the message was returned because it couldn't be decrypted, write it to the encrypted messages directory
        if ( hasattr(retval,'key') and hasattr(retval,'value') and
             (isinstance(retval.key,KafkaCryptoMessage) or isinstance(retval.value,KafkaCryptoMessage)) ) :
            if not self.__encrypted_messages_subdir.is_dir() :
                self.__encrypted_messages_subdir.mkdir(parents=True)
            key_fn, value_fn = get_encrypted_message_key_and_value_filenames(retval,self.topic_name)
            key_fp = self.__encrypted_messages_subdir/key_fn
            value_fp = self.__encrypted_messages_subdir/value_fn
            warnmsg = 'WARNING: encountered a message that failed to be decrypted. Key bytes will be written to '
            warnmsg+= f'{key_fp.relative_to(self.dirpath)} and value bytes will be written to '
            warnmsg+= f'{value_fp.relative_to(self.dirpath)}'
            self.logger.warning(warnmsg)
            with open(key_fp,'wb') as fp :
                fp.write(bytes(retval.key))
            with open(value_fp,'wb') as fp :
                fp.write(bytes(retval.value))
            return False #because the message wasn't processed successfully
        if retval is True :
            return retval
        #get the DataFileChunk from the message value
        try :
            dfc = msg.value() #from a regular Kafka Consumer
        except TypeError :
            dfc = msg.value #from KafkaCrypto
        #If the file was successfully reconstructed, return True
        if retval==DATA_FILE_HANDLING_CONST.FILE_SUCCESSFULLY_RECONSTRUCTED_CODE :
            infomsg = f'File {self.files_in_progress_by_path[dfc.filepath].full_filepath.relative_to(dfc.rootdir)} '
            infomsg+= 'successfully reconstructed from stream'
            self.logger.info(infomsg)
            self.completely_processed_filepaths.append(dfc.filepath)
            with lock :
                del self.files_in_progress_by_path[dfc.filepath]
                del self.locks_by_fp[dfc.filepath]
            return True
        #If the file hash was mismatched after reconstruction, return False
        if retval==DATA_FILE_HANDLING_CONST.FILE_HASH_MISMATCH_CODE :
            warnmsg = f'WARNING: hashes for file {self.files_in_progress_by_path[dfc.filepath].filename} not matched '
            warnmsg+= 'after reconstruction! All data have been written to disk, but not as they were uploaded.'
            self.logger.warning(warnmsg)
            with lock :
                del self.files_in_progress_by_path[dfc.filepath]
                del self.locks_by_fp[dfc.filepath]
            return False
        #if this is reached the return code was unrecognized
        self.logger.error(f'ERROR: unrecognized add_chunk return value ({retval})!',NotImplementedError)
        return False

    def _on_check(self) :
        msg = f'{self.n_msgs_read} messages read, {self.n_msgs_processed} messages processed, '
        msg+= f'{len(self.completely_processed_filepaths)} files completely reconstructed so far'
        self.logger.debug(msg)
        if len(self.files_in_progress_by_path)>0 or len(self.completely_processed_filepaths)>0 :
            self.logger.debug(self.progress_msg)

    #################### CLASS METHODS ####################

    @classmethod
    def get_command_line_arguments(cls) :
        superargs,superkwargs = super().get_command_line_arguments()
        args = [*superargs,'output_dir','config','topic_name','update_seconds','consumer_group_id']
        kwargs = {**superkwargs,'n_threads':RUN_OPT_CONST.N_DEFAULT_DOWNLOAD_THREADS}
        return args,kwargs

    @classmethod
    def run_from_command_line(cls,args=None) :
        """
        Run a :class:`~DataFileDownloadDirectory` directly from the command line

        Calls :func:`~reconstruct` on a :class:`~DataFileDownloadDirectory` defined by
        command line (or given) arguments

        :param args: the list of arguments to send to the parser instead of getting them from sys.argv
        :type args: list, optional
        """
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        #make the download directory
        reconstructor_directory = cls(args.output_dir,args.config,args.topic_name,
                                      n_threads=args.n_threads,
                                      consumer_group_id=args.consumer_group_id,
                                      update_secs=args.update_seconds,
                                     )
        #start the reconstructor running
        run_start = datetime.datetime.now()
        reconstructor_directory.logger.info(f'Listening for files to reconstruct in {args.output_dir}')
        n_read,n_processed,complete_filenames = reconstructor_directory.reconstruct()
        reconstructor_directory.close()
        run_stop = datetime.datetime.now()
        #shut down when that function returns
        reconstructor_directory.logger.info(f'File reconstructor writing to {args.output_dir} shut down')
        msg = f'{n_read} total messages were consumed'
        if len(complete_filenames)>0 :
            msg+=f', {n_processed} messages were successfully processed,'
            msg+=f' and the following {len(complete_filenames)} file'
            msg+=' was' if len(complete_filenames)==1 else 's were'
            msg+=' successfully reconstructed'
            for fn in complete_filenames :
                msg+=f'\n\t{fn}'
        else :
            msg+=f' and {n_processed} messages were successfully processed'
        msg+=f' from {run_start} to {run_stop}'
        reconstructor_directory.logger.info(msg)

def main(args=None) :
    """
    Main method to run from command line
    """
    DataFileDownloadDirectory.run_from_command_line(args)

if __name__=='__main__' :
    main()



"""
A utility class to represent a set of (relatively arbitrary) dataclass objects serialized to/deseralized from
a corresponding CSV file in an atomic and (relatively) thread-safe way
"""

#imports
import pathlib, functools, datetime, typing, copy, os
from threading import Lock
from dataclasses import fields, is_dataclass
import methodtools
from atomicwrites import atomic_write
from .logging import LogOwner

def get_nested_types() :
    """
    Helper function to generate the class constant that holds the supported nested types for dataclass tables
    """
    simple_types = [str,int,float,complex,bool]
    container_types = [(typing.List,list,'[]'),(typing.Tuple,tuple,'()'),(typing.Set,set,'set()')]
    nested_types_dict = {}
    for c_t in container_types :
        for s_t in simple_types :
            nested_types_dict[c_t[0][s_t]] = (c_t[1],s_t,c_t[2])
    return nested_types_dict

class DataclassTable(LogOwner) :
    """
    A class to work with an atomic csv file that's holding dataclass entries in a thread-safe way

    :param dataclass_type: The :class:`dataclasses.dataclass` defining the entries in the table/csv file
    :type dataclass_type: :class:`dataclasses.dataclass`
    :param filepath: The path to the .csv file that should be created (or read from) on startup.
        The default is a file named after the dataclass type in the current directory.
    :type filepath: :class:`pathlib.Path` or None, optional
    """

    #################### PROPERTIES AND CONSTANTS ####################

    DELIMETER = ';' #can't use a comma or containers would display incorrectly
    DATETIME_FORMAT = '%a %b %d, %Y at %H:%M:%S'
    THREAD_LOCK = Lock()
    UPDATE_FILE_EVERY = 5 #only update the .csv file automatically every 5 seconds to make updates less expensive

    NESTED_TYPES = get_nested_types()

    @methodtools.lru_cache()
    @property
    def csv_header_line(self) :
        """
        The first line in the CSV file (OS-dependent)
        """
        header_line = ''
        #add an extra line when running on Windows to seamlessly open the file in Excel
        if os.name=='nt' :
            header_line+=f'sep={DataclassTable.DELIMETER}\n'
        for fieldname in self.__field_names :
            header_line+=f'{fieldname}{DataclassTable.DELIMETER}'
        return header_line[:-1]

    @property
    def obj_addresses(self) :
        """
        All recognized object hex addresses
        """
        return list(self.__entry_objs.keys())

    @property
    def filepath(self) :
        """
        Path to the CSV file
        """
        return self.__filepath

    @property
    def lock(self) :
        """
        A thread lock to use for ensuring only one thread is interacting with the :class:`~DataclassTable` at a time
        """
        return DataclassTable.THREAD_LOCK

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,dataclass_type,*,filepath=None,**kwargs) :
        """
        Constructor method
        """
        #init the LogOwner
        super().__init__(**kwargs)
        #figure out what type of objects the table/file will be describing
        self.__dataclass_type = dataclass_type
        if not is_dataclass(self.__dataclass_type) :
            self.logger.error(f'ERROR: "{self.__dataclass_type}" is not a dataclass!',TypeError)
        if len(fields(self.__dataclass_type)) <= 0 :
            errmsg = f'ERROR: dataclass type {self.__dataclass_type} does not have any fields '
            errmsg+= f'and so cannot be used in a {self.__class__.__name__}!'
            self.logger.error(errmsg,ValueError)
        self.__field_names = [field.name for field in fields(self.__dataclass_type)]
        self.__field_types = [field.type for field in fields(self.__dataclass_type)]
        #figure out where the csv file should go
        self.__filepath = filepath if filepath is not None else pathlib.Path() / f'{self.__dataclass_type.__name__}.csv'
        #set some other variables
        self.__entry_objs = {}
        self.__entry_lines = {}
        #read or create the file to finish setting up the table
        self.__file_last_updated = datetime.datetime.now()
        if self.__filepath.is_file() :
            self.__read_csv_file()
        else :
            msg = f'Creating new {self.__class__.__name__} csv file at {self.__filepath} '
            msg+= f'to hold {self.__dataclass_type.__name__} entries'
            self.logger.info(msg)
            self.dump_to_file()

    def __del__(self) :
        self.dump_to_file(reraise_exc=False)

    def add_entries(self,new_entries) :
        """
        Add a new set of entries to the table.

        :param new_entries: the new entry or entries to add to the table
        :type new_entries: :class:`dataclasses.dataclass` or list(:class:`dataclasses.dataclass`)

        :raises ValueError: if any of the objects in `new_entries` already exists in the table
        """
        if is_dataclass(new_entries) :
            new_entries = [new_entries]
        for entry in new_entries :
            entry_addr = hex(id(entry))
            if entry_addr in self.__entry_objs or entry_addr in self.__entry_lines :
                self.logger.error('ERROR: address of object sent to add_entries is already registered!',ValueError)
            with DataclassTable.THREAD_LOCK :
                self.__entry_objs[entry_addr] = entry
                self.__entry_lines[entry_addr] = self.__line_from_obj(entry)
                self.obj_addresses_by_key_attr.cache_clear()
        if (datetime.datetime.now()-self.__file_last_updated).total_seconds()>DataclassTable.UPDATE_FILE_EVERY :
            self.dump_to_file()

    def remove_entries(self,entry_obj_addresses) :
        """
        Remove an entry or entries from the table

        :param entry_obj_addresses: a single value or container of entry addresses (object IDs in hex form) to remove
            from the table
        :type entry_obj_addresses: hex(id(object)) or list(hex(id(object)))

        :raises ValueError: if any of the hex addresses in `entry_obj_addresses` is not present in the table
        """
        if isinstance(entry_obj_addresses,str) :
            entry_obj_addresses = [entry_obj_addresses]
        for entry_addr in entry_obj_addresses :
            if (entry_addr not in self.__entry_objs) or (entry_addr not in self.__entry_lines) :
                self.logger.error(f'ERROR: address {entry_addr} sent to remove_entries is not registered!',ValueError)
            with DataclassTable.THREAD_LOCK :
                self.__entry_objs.pop(entry_addr)
                self.__entry_lines.pop(entry_addr)
                self.obj_addresses_by_key_attr.cache_clear()
        if (datetime.datetime.now()-self.__file_last_updated).total_seconds()>DataclassTable.UPDATE_FILE_EVERY :
            self.dump_to_file()

    def get_entry_attrs(self,entry_obj_address,*args) :
        """
        Return copies of all or some of the current attributes of an entry in the table.
        Returning copies ensures the original objects cannot be modified by accident.

        Use `args` to get a dictionary of desired attribute values returned.
        If only one arg is given the return value is just that single attribute.
        The default (no additional arguments) returns a dictionary of all attributes for the entry

        :param entry_obj_address: the address in memory of the object to return copies of attributes for
        :type entry_obj_address: hex(id(object))
        :param args: Add other arguments that are names of attributes to get only those specific attributes of the entry
        :type args: str, optional

        :return: copies of some or all attributes for an entry in the table
        :rtype: depends on arguments, or dict

        :raises ValueError: if `entry_obj_address` doesn't correspond to an object listed in the table
        """
        if entry_obj_address not in self.__entry_objs :
            errmsg = f'ERROR: address {entry_obj_address} sent to get_entry_attrs is not registered!'
            self.logger.error(errmsg,ValueError)
        obj = self.__entry_objs[entry_obj_address]
        if len(args)==1 :
            return copy.deepcopy(getattr(obj,args[0]))
        to_return = {}
        for fname in self.__field_names :
            if (not args) or (fname in args) :
                to_return[fname] = copy.deepcopy(getattr(obj,fname))
        if args :
            for arg in args :
                if arg not in to_return :
                    errmsg = f'WARNING: attribute name {arg} is not a name of a {self.__dataclass_type} '
                    errmsg+= 'Field and will not be returned from get_entry_attrs!'
                    self.logger.warning(errmsg)
        return to_return

    def set_entry_attrs(self,entry_obj_address,**kwargs) :
        """
        Modify attributes of an entry that already exists in the table

        :param entry_obj_address: The address in memory of the entry object to modify
        :type entry_obj_address: hex(id(object))
        :param kwargs: Attributes to set (keys are names, values are values for those named attrs)
        :type kwargs: dict

        :raises ValueError: if `entry_obj_address` doesn't correspond to an object listed in the table
        """
        if entry_obj_address not in self.__entry_objs :
            errmsg = f'ERROR: address {entry_obj_address} sent to set_entry_attrs is not registered!'
            self.logger.error(errmsg,ValueError)
        with DataclassTable.THREAD_LOCK :
            obj = self.__entry_objs[entry_obj_address]
            for fname,fval in kwargs.items() :
                setattr(obj,fname,fval)
            self.__entry_lines[entry_obj_address] = self.__line_from_obj(obj)
            self.obj_addresses_by_key_attr.cache_clear()
        if (datetime.datetime.now()-self.__file_last_updated).total_seconds()>DataclassTable.UPDATE_FILE_EVERY :
            self.dump_to_file()

    @functools.lru_cache(maxsize=8)
    def obj_addresses_by_key_attr(self,key_attr_name) :
        """
        Return a dictionary whose keys are the values of some given attribute for each object
        and whose values are lists of the addresses in memory of the objects in the table
        that have each value of the requested attribute.

        Useful to find objects in the table by attribute values so they can be efficiently updated
        without compromising the integrity of the objects in the table and their attributes.

        Up to five calls are cached so if nothing changes this happens a little faster.

        :param key_attr_name: the name of the attribute whose values should be used as keys in the returned dictionary
        :type key_attr_name: str

        :return: A dictionary listing all objects in the table, keyed by their values of `key_attr_name`
        :rtype: dict

        :raises ValueError: if `key_attr_name` is not recognized as the name of an attribute for entries in the table
        """
        if key_attr_name not in self.__field_names :
            errmsg = f'ERROR: {key_attr_name} is not a name of a Field for {self.__dataclass_type} objects!'
            self.logger.error(errmsg,ValueError)
        to_return = {}
        with DataclassTable.THREAD_LOCK :
            for addr,obj in self.__entry_objs.items() :
                rkey = getattr(obj,key_attr_name)
                if rkey not in to_return :
                    to_return[rkey] = []
                to_return[rkey].append(addr)
        return to_return

    def dump_to_file(self,reraise_exc=True) :
        """
        Dump the contents of the table to a csv file.
        Call this to force the file to update and reflect the current state of objects.
        Automatically called in several contexts.
        """
        lines_to_write = [self.csv_header_line]
        if len(self.__entry_lines)>0 :
            lines_to_write+=list(self.__entry_lines.values())
        with DataclassTable.THREAD_LOCK :
            self.__write_lines(lines_to_write,reraise_exc=reraise_exc)

    #################### PRIVATE HELPER FUNCTIONS ####################

    def __read_csv_file(self) :
        """
        Read in a csv file with lines of the expected dataclass type
        """
        #read the file
        with open(self.__filepath,'r') as fp :
            lines_as_read = fp.readlines()
        #first make sure the header line matches what we'd expect for the dataclass
        n_header_lines = len(self.csv_header_line.split('\n'))
        for ihl in range(n_header_lines) :
            if lines_as_read[ihl].strip()!=(self.csv_header_line.split('\n'))[ihl] :
                errmsg = f'ERROR: header line in {self.__filepath} ({lines_as_read[0]}) does not match expectation for '
                errmsg+= f'{self.__dataclass_type} ({self.csv_header_line})!'
                self.logger.error(errmsg,RuntimeError)
        #add entry lines and objects
        for line in lines_as_read[n_header_lines:] :
            obj = self.__obj_from_line(line)
            #key both dictionaries by the address of the object in memory
            dkey = hex(id(obj))
            self.__entry_objs[dkey] = obj
            self.__entry_lines[dkey] = line
        msg = f'Found {len(self.__entry_objs)} {self.__dataclass_type.__name__} entries in {self.__filepath}'
        self.logger.info(msg)

    def __write_lines(self,lines,overwrite=True,reraise_exc=True) :
        """
        Write a line or container of lines to the csv file, in a thread-safe and atomic way
        """
        if isinstance(lines,str) :
            lines = [lines]
        lines_string = ''
        for line in lines :
            lines_string+=f'{line.strip()}\n'
        try :
            with atomic_write(self.__filepath,overwrite=overwrite) as fp :
                fp.write(lines_string)
        except Exception as exc :
            if not isinstance(exc,FileNotFoundError) : #This is common to see and okay when running multithreaded
                if reraise_exc :
                    errmsg = f'ERROR: failed to write to {self.__class__.__name__} csv file at {self.__filepath}! '
                    errmsg+=  'Will reraise exception.'
                    self.logger.error(errmsg,exc_obj=exc)
                else :
                    msg = f'WARNING: failed an attempt to write to {self.__class__.__name__} csv file at '
                    msg+= f'{self.__filepath}! Exception ({type(exc)}): {exc}'
                    self.logger.warning(msg)
        self.__file_last_updated = datetime.datetime.now()

    def __line_from_obj(self,obj) :
        """
        Return the csv file line for a given object
        """
        if obj.__class__ != self.__dataclass_type :
            self.logger.error(f'ERROR: "{obj}" is mismatched to type {self.__dataclass_type}!',TypeError)
        obj_line = ''
        for fname,ftype in zip(self.__field_names,self.__field_types) :
            obj_line+=f'{self.__get_str_from_attribute(getattr(obj,fname),ftype)}{DataclassTable.DELIMETER}'
        return obj_line[:-1]

    def __obj_from_line(self,line) :
        """
        Return the dataclass instance for a given csv file line string
        """
        args = []
        for attrtype,attrstr in zip(self.__field_types,(line.strip().split(DataclassTable.DELIMETER))) :
            args.append(self.__get_attribute_from_str(attrstr,attrtype))
        return self.__dataclass_type(*args)

    def __get_str_from_attribute(self,attrobj,attrtype) :
        """
        Given an object and the type it is in the dataclass,
        return the string representation of it that should go in the file
        """
        if attrtype==datetime.datetime :
            return repr(attrobj.strftime(DataclassTable.DATETIME_FORMAT))
        if attrtype==pathlib.Path :
            return repr(str(attrobj))
        return repr(attrobj)

    def __get_attribute_from_str(self,attrstr,attrtype) :
        """
        Given the string of the repr() output of some object and the datatype it represents, return the actual object
        """
        #datetime objects are handled in a custom way
        if attrtype==datetime.datetime :
            return datetime.datetime.strptime(attrstr[1:-1],DataclassTable.DATETIME_FORMAT)
        #so are path objects
        if attrtype==pathlib.Path :
            return pathlib.Path(attrstr[1:-1])
        #strings have extra quotes on either end
        if attrtype==str :
            return attrtype(attrstr[1:-1])
        #int, float, complex, and bool can all be directly re-casted
        if attrtype in (int,float,complex,bool) :
            return attrtype(attrstr)
        #some simply-nested container types can be casted in two steps
        if attrtype in self.NESTED_TYPES :
            #empty collections
            if attrstr==self.NESTED_TYPES[attrtype][2] :
                return self.NESTED_TYPES[attrtype][0]()
            to_cast = []
            for vstr in attrstr[1:-1].split(',') :
                to_cast.append(self.NESTED_TYPES[attrtype][1](vstr))
            return self.NESTED_TYPES[attrtype][0](to_cast)
        errmsg = f'ERROR: attribute type "{attrtype}" is not recognized for a {self.__class__.__name__}!'
        self.logger.error(errmsg,ValueError)
        return None



"""Base class for messages computed from reconstructed data files and produced to new topics"""

#imports
from abc import ABC
from ...kafka_wrapper.producible import Producible
from ..utilities import get_message_prepend

class ReproducerMessage(Producible,ABC) :
    """
    Abstract base class for messages that are computed from data files and produced to a new topic

    :param datafile: The DataFile object used to compute this message
    :type datafile: :class:`~.data_file_io.DownloadDataFileToMemory`
    """

    @property
    def msg_key(self) :
        """
        The automatically defined key for these messages is a string version of the path
        to the original file with "_processing_result" appended
        """
        key_pp = get_message_prepend(self.__datafile.subdir_str,self.__datafile.filename)
        return f'{key_pp}_processing_result'

    @property
    def callback_kwargs(self) :
        """
        The callback kwargs defined for these messages are the original datafile name, its relative filepath,
        and the total number of chunks into which the file was broken.
        """
        if self.__datafile.subdir_str!='' :
            rel_filepath = f'{self.__datafile.subdir_str}/{self.__datafile.filename}'
        else :
            rel_filepath = self.__datafile.filename
        return {'filename':self.__datafile.filename,
                'rel_filepath':rel_filepath,
                'n_total_chunks':self.__datafile.n_total_chunks}

    def get_log_msg(self,print_every=None) :
        """
        Prints a message saying that the processing result from [filepath] is being produced, if print_every is defined
        """
        if print_every :
            return f'Producing processing result from {self.__datafile.filepath}'
        return None

    def __init__(self,datafile,*args,**kwargs) :
        super().__init__(*args,**kwargs)
        self.__datafile = datafile



"""Manage Linux daemons"""

#imports
import sys, os, pathlib, textwrap
from .config import SERVICE_CONST
from .utilities import run_cmd_in_subprocess
from .service_manager_base import ServiceManagerBase

class LinuxServiceManager(ServiceManagerBase) :
    """
    Base class for working with Linux daemons

    :param service_name: The name of the daemon as installed
    :type service_name: str
    :param service_spec_string: A string specifying which code should be run as a daemon.
        Could be the name of an OpenMSIStream Runnable class, or the path to a custom Python code.
        Custom Services can also specify a :class:`~.workflow.Runnable` class name,
        and/or a function in the file using special formatting like [class_name]=[path.to.file]:[function_name].
        Only needed to initially install the daemon.
    :type service_spec_string: str, optional
    :param argslist: The list of arguments (as from the command line) to pass to the
        :class:`~.workflow.Runnable` class.
        Only needed to initially install the daemon.
    :type argslist: list, optional
    :param interactive: if True, a few more messages/prompts will come up telling a user what to do
    :type interactive: bool, optional
    """

    @property
    def env_var_names(self):
        for evn in super().env_var_names :
            yield evn
        #get the names of environment variables from the env_var file
        if self.env_var_filepath.is_file() :
            with open(self.env_var_filepath,'r') as fp :
                lines = fp.readlines()
            for line in lines :
                linesplit = (line.strip()).split('=')
                if len(linesplit)==2 :
                    yield linesplit[0]

    def __init__(self,*args,**kwargs) :
        super().__init__(*args,**kwargs)
        self.daemon_working_dir_filepath = SERVICE_CONST.WORKING_DIR/f'{self.service_name}.service'
        self.daemon_filepath = SERVICE_CONST.DAEMON_SERVICE_DIR/self.daemon_working_dir_filepath.name

    def install_service(self) :
        """
        Install the daemon
        """
        super().install_service()
        #make sure systemd is running
        self.__check_systemd_installed()
        #write the daemon file pointing to the executable
        self.__write_daemon_file()
        #enable the service
        run_cmd_in_subprocess(['sudo','systemctl','daemon-reload'],logger=self.logger)
        run_cmd_in_subprocess(['sudo','systemctl','enable',f'{self.service_name}.service'],logger=self.logger)
        self.logger.info(f'Done installing {self.service_name}')

    def start_service(self) :
        """
        Start the daemon
        """
        self.logger.info(f'Starting {self.service_name}...')
        self.__check_systemd_installed()
        run_cmd_in_subprocess(['sudo','systemctl','daemon-reload'],logger=self.logger)
        run_cmd_in_subprocess(['sudo','systemctl','start',f'{self.service_name}.service'],logger=self.logger)
        self.logger.info(f'Done starting {self.service_name}')

    def service_status(self) :
        """
        Print the status of the daemon
        """
        self.__check_systemd_installed()
        result = run_cmd_in_subprocess(['sudo','systemctl','status',f'{self.service_name}.service'],logger=self.logger)
        self.logger.info(f'{self.service_name} status: {result.decode()}')

    def stop_service(self) :
        """
        Stop the daemon
        """
        self.logger.info(f'Stopping {self.service_name}...')
        self.__check_systemd_installed()
        run_cmd_in_subprocess(['sudo','systemctl','stop',f'{self.service_name}.service'],logger=self.logger)
        self.logger.info(f'Done stopping {self.service_name}')

    def remove_service(self,remove_env_vars=False,remove_install_args=False,remove_nssm=False) :
        """
        Remove the daemon.

        :param remove_env_vars: if True, any environment variables needed by the daemon will be removed.
        :type remove_env_vars: bool, optional.
        :param remove_install_args: if True, the file listing the arguments used to install the daemon
            (to make it easier to re-install) will be removed.
        :type remove_install_args: bool, optional
        :param remove_nssm: Not used for Linux daemons.
        :type remove_nssm: bool, optional
        """
        self.logger.info(f'Removing {self.service_name}...')
        self.__check_systemd_installed()
        try :
            run_cmd_in_subprocess(['sudo','systemctl','disable',f'{self.service_name}.service'],logger=self.logger)
        except Exception :
            pass
        if self.daemon_filepath.exists() :
            run_cmd_in_subprocess(['sudo','rm','-f',str(self.daemon_filepath)],logger=self.logger)
        run_cmd_in_subprocess(['sudo','systemctl','daemon-reload'],logger=self.logger)
        run_cmd_in_subprocess(['sudo','systemctl','reset-failed'],logger=self.logger)
        self.logger.info('Service removed')
        if remove_nssm :
            warnmsg = "WARNING: requested to remove NSSM along with the Service, "
            warnmsg+= "but Linux Services don't install NSSM so nothing was done."
            self.logger.warning(warnmsg)
        super().remove_service(remove_env_vars=remove_env_vars,remove_install_args=remove_install_args)

    def _write_env_var_file(self) :
        code = ''
        for evn in self.env_var_names :
            val = os.path.expandvars(f'${evn}')
            if val==f'${evn}' :
                raise RuntimeError(f'ERROR: value not found for expected environment variable {evn}!')
            code += f'{evn}={val}\n'
        if code=='' :
            return False
        with open(self.env_var_filepath,'w') as fp :
            fp.write(code)
        run_cmd_in_subprocess(['chmod','go-rwx',self.env_var_filepath])
        return True

    def __check_systemd_installed(self) :
        """
        Raises an error if systemd is not installed (systemd is needed to control Linux daemons)
        """
        check = run_cmd_in_subprocess(['ps','--no-headers','-o','comm','1'],logger=self.logger)
        if check.decode().rstrip()!='systemd' :
            errmsg = 'ERROR: Installing programs as Services ("daemons") on Linux requires systemd!'
            errmsg = 'You can install systemd with "sudo apt install systemd" (or similar) and try again.'
            self.logger.error(errmsg,RuntimeError)

    def __write_daemon_file(self) :
        """
        Write the Unit/.service file to the daemon directory that calls the Python executable
        """
        #make sure the directory to hold the file exists
        if not SERVICE_CONST.DAEMON_SERVICE_DIR.is_dir() :
            SERVICE_CONST.logger.info(f'Creating a new daemon service directory at {SERVICE_CONST.DAEMON_SERVICE_DIR}')
            SERVICE_CONST.DAEMON_SERVICE_DIR.mkdir(parents=True)
        #write out the file pointing to the python executable
        description = 'Python script'
        if self.service_dict['class'] is not None :
            classdoc = self.service_dict['class'].__doc__
            if classdoc is not None :
                description = classdoc.strip().replace('\n',' ')
            else :
                description+=f" for a {self.service_dict['class']} object"
        code = f'''\
            [Unit]
            Description = {description}
            Requires = network-online.target remote-fs.target
            After = network-online.target remote-fs.target

            [Service]
            Type = simple
            User = {os.path.expandvars('$USER')}
            ExecStart = {sys.executable} {self.exec_fp}'''
        if self.env_vars_needed :
            code+=f'''\n\
            EnvironmentFile = {self.env_var_filepath}'''
        code+=f'''\n\
            WorkingDirectory = {pathlib.Path().resolve()}
            Restart = on-failure
            RestartSec = 30

            [Install]
            WantedBy = multi-user.target'''
        with open(self.daemon_working_dir_filepath,'w') as fp :
            fp.write(textwrap.dedent(code))
        run_cmd_in_subprocess(['sudo','mv',str(self.daemon_working_dir_filepath),str(self.daemon_filepath.parent)],
                              logger=self.logger)



"""A single data file"""

#imports
from ...utilities import LogOwner
from ...utilities.misc import populated_kwargs

class DataFile(LogOwner) :
    """
    Base class for representing a single data file

    :param filepath: the path to the file
    :type filepath: :class:`pathlib.Path`
    """

    def __init__(self,filepath,*args,**kwargs) :
        self.filepath = filepath
        self.filename = self.filepath.name
        kwargs = populated_kwargs(kwargs,{'logger_file':self.filepath.parent})
        super().__init__(*args,**kwargs)



"""
A stream handler that triggers a function to compute a message on file completion.
Produces the computed message to a different topic.
"""

#imports
from abc import ABC, abstractmethod
from ..config import DATA_FILE_HANDLING_CONST
from .data_file_chunk_handlers import DataFileChunkReproducer
from .data_file_stream_handler import DataFileStreamHandler
from .file_registry.stream_handler_registries import StreamReproducerRegistry

class DataFileStreamReproducer(DataFileStreamHandler,DataFileChunkReproducer,ABC) :
    """
    A class to consume :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk` messages into memory,
    compute some processing result when entire files are available, and produce that result to a different topic.

    This is a base class that cannot be instantiated on its own.

    :param config_path: Path to the config file to use in defining the Broker connection, Consumers, and Producers
    :type config_path: :class:`pathlib.Path`
    :param consumer_topic_name: Name of the topic to which the Consumers should be subscribed
    :type consumer_topic_name: str
    :param producer_topic_name: Name of the topic to which the Producer should produce the processing results
    :type producer_topic_name: str
    :param output_dir: Path to the directory where the log and csv registry files should be kept (if None a default
        will be created in the current directory)
    :type output_dir: :class:`pathlib.Path`, optional
    :param datafile_type: the type of data file that recognized files should be reconstructed as
        (must be a subclass of :class:`~.data_file_io.DownloadDataFileToMemory`)
    :type datafile_type: :class:`~.data_file_io.DownloadDataFileToMemory`, optional
    :param n_producer_threads: the number of producers to run. The total number of producer/consumer threads
        started is `max(n_consumer_threads,n_producer_threads)`.
    :type n_producer_threads: int, optional
    :param n_consumer_threads: the number of consumers to run. The total number of producer/consumer threads
        started is `max(n_consumer_threads,n_producer_threads)`.
    :type n_consumer_threads: int, optional
    :param consumer_group_id: the group ID under which each consumer should be created
    :type consumer_group_id: str, optional

    :raises ValueError: if `datafile_type` is not a subclass of
        :class:`~.data_file_io.DownloadDataFileToMemory`
    """

    def __init__(self,config_file,consumer_topic_name,producer_topic_name,**kwargs) :
        """
        Constructor method signature duplicated above to display in Sphinx docs
        """
        super().__init__(config_file,consumer_topic_name,producer_topic_name,**kwargs)

    def produce_processing_results_for_files_as_read(self) :
        """
        Consumes messages in several parallel threads and stores their data in memory. Calls
        :func:`~_get_processing_result_message_for_file` for fully-read files to get their
        processing result messages. Produces processing result messages in several parallel
        threads as they're generated. Runs until the user inputs a command to shut it down.

        :return: the total number of messages consumed
        :rtype: int
        :return: the total number of messages processed (registered in memory)
        :rtype: int
        :return: the paths of files successfully reconstructed from the Consumer topic during the run
        :rtype: list
        :return: the paths of files whose processing results were successfully produced to the Producer topic
            during the run
        :rtype: list
        """
        #startup message
        msg = f'Will process files from messages in the {self.consumer_topic_name} topic using '
        msg+= f'{self.n_consumer_threads} thread{"s" if self.n_consumer_threads>1 else ""} and produce their '
        msg+= f'processing results to the {self.producer_topic_name} topic using {self.n_producer_threads} '
        msg+= f'thread{"s" if self.n_producer_threads>1 else ""}'
        self.logger.info(msg)
        #set up the stream reproducer registry
        self.file_registry = StreamReproducerRegistry(dirpath=self._output_dir,
                                                       consumer_topic_name=self.consumer_topic_name,
                                                       consumer_group_id=self.consumer_group_id,
                                                       producer_topic_name=self.producer_topic_name,
                                                       logger=self.logger)
        #if there are files that need to be re-processed, set the variables to re-read messages from those files
        if self.file_registry.rerun_file_key_regex is not None :
            msg = f'Consumer{"s" if self.n_consumer_threads>1 else ""} will start from the beginning of the topic to '
            msg+= f're-read messages for {self.file_registry.n_files_to_rerun} previously-failed '
            msg+= f'file{"s" if self.file_registry.n_files_to_rerun>1 else ""}'
            self.logger.info(msg)
            self.restart_at_beginning=True
            self.message_key_regex=self.file_registry.rerun_file_key_regex
        #create the arguments for each _run_worker thread
        run_worker_args_per_thread = []
        for ti in range(self.n_threads) :
            run_worker_args_per_thread.append([ti<self.n_consumer_threads,ti<self.n_producer_threads])
        run_worker_kwargs_per_thread = {'produce_from_queue_kwargs':{'callback':self.producer_callback,'print_every':1}}
        #call the run loop
        self.run(args_per_thread=run_worker_args_per_thread,kwargs_per_thread=run_worker_kwargs_per_thread)
        #return the results of the processing
        return ( self.n_msgs_read, self.n_msgs_processed,
                 self.completely_processed_filepaths, self.results_produced_filepaths )

    def producer_callback(self,err,msg,filename,rel_filepath,n_total_chunks) :
        """
        A reference to this method is given as the callback for each call to :func:`confluent_kafka.Producer.produce`.
        It is called for every message upon acknowledgement by the broker, and it uses the file registries in the
        LOGS subdirectory to keep the information about what has and hasn't been uploaded current with what has
        been received by the broker.

        Messages associated with an error from the broker will be recomputed and added back to the queue
        to be produced again, logging an error and registering the file as failed if the message can't be
        computed from the datafile.

        Messages that are successfully produced will move their associated data files to the "results_produced"
        csv file.

        :param err: The error object for the message
        :type err: :class:`confluent_kafka.KafkaError`
        :param msg: The message object
        :type msg: :class:`confluent_kafka.Message`
        :param filename: The name of the file that was used to create this processing result message
        :type filename: str
        :param rel_filepath: The path to the file that was used to create this processing result message,
            relative to the :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk`'s root directory
        :type rel_filepath: :class:`pathlib.Path`
        :param n_total_chunks: The total number of chunks in the file used to create this processing result message
        :type n_total_chunks: int
        """
        # If no error occured, increment the counter for the number of messages produced
        if err is None and msg.error() is None :
            with self.lock :
                self.n_msgs_produced+=1
        # If any error occured, log a warning and re-enqueue the message to be produced again
        if err is None and msg.error() is not None :
            err = msg.error()
        if err is not None :
            self.file_registry.register_file_result_production_failed(filename,rel_filepath,n_total_chunks)
            if err.fatal() :
                warnmsg =f'WARNING: fatally failed to deliver processing result message for {rel_filepath}. '
                warnmsg+=f'This message will be re-enqueued. Error reason: {err.str()}'
            elif not err.retriable() :
                warnmsg =f'WARNING: Failed to deliver processing result message for {rel_filepath} and cannot retry. '
                warnmsg+= f'This message will be re-enqueued. Error reason: {err.str()}'
            self.logger.warning(warnmsg)
            datafile = self.files_in_progress_by_path[self._output_dir/rel_filepath]
            new_msg = self._get_processing_result_message_for_file(datafile,self.lock)
            if new_msg is not None :
                self.producer_message_queue.put(new_msg)
            else :
                self._failed_computing_processing_result(datafile,self.lock)
        # Otherwise, register the associated file as having its processing result successfully produced
        else :
            with self.lock :
                self.file_registry.register_file_results_produced(filename,rel_filepath,n_total_chunks)
                self.results_produced_filepaths.append(self._output_dir/rel_filepath)
                #stop tracking the file
                del self.files_in_progress_by_path[self._output_dir/rel_filepath]
                del self.locks_by_fp[self._output_dir/rel_filepath]
            infomsg = f'Processing result message for {rel_filepath} has been received by the broker for '
            infomsg+= f'the {self.producer_topic_name} topic'
            self.logger.info(infomsg)

    def _process_message(self,lock,msg,rootdir_to_set=None):
        """
        Process a single message to add it to a file being held in memory until all messages are received.

        If the message failed to be decrypted, this method calls :func:`~_undecryptable_message_callback` and returns.

        If the message is the first one consumed for a particular file, or any message other than the last one needed,
        it registers the file as 'in_progress' in the .csv file.

        If the message is the last message needed for a file and its contents match the original hash
        of the file on disk, this method calls :func:`~_get_processing_result_message_for_file` and enqueues
        the result to be produced. The file is moved to the 'results_produced' .csv file when the produced
        message is acknowledged by the broker through :func:`~producer_callback`.

        If the call to :func:`~_get_processing_result_message_for_file` returns None, this method calls
        :func:`~_failed_computing_processing_result` and returns.

        If the message is the last one needed but the contents are somehow different than the original file on disk,
        this method calls :func:`~_mismatched_hash_callback`, registers the file as 'mismatched_hash' in the .csv file,
        and returns.

        :param lock: Acquiring this :class:`threading.Lock` object ensures that only one instance
            of :func:`~_process_message` is running at once
        :type lock: :class:`threading.Lock`
        :param msg: The received :class:`confluent_kafka.KafkaMessage` object, or an undecrypted KafkaCrypto message
        :type msg: :class:`confluent_kafka.KafkaMessage` or :class:`kafkacrypto.Message`
        :param rootdir_to_set: Path to a directory that should be set as the "root" for reconstructed data files
            (default is the output directory)
        :type rootdir_to_set: :class:`pathlib.Path`

        :return: True if processing the message was successful (file in progress or message enqueue to be produced),
            False otherwise
        :rtype: bool
        """
        retval = super()._process_message(lock,msg,self._output_dir if rootdir_to_set is None else rootdir_to_set)
        #if the file was in progress or had a mismatched hash, return True or False, respectively
        if retval in (True,False) :
            return retval
        #get the DataFileChunk from the message value
        try :
            dfc = msg.value() #from a regular Kafka Consumer
        except TypeError :
            dfc = msg.value #from KafkaCrypto
        #if the file has had all of its messages read successfully, try to enqueue its processing result to produce
        if retval==DATA_FILE_HANDLING_CONST.FILE_SUCCESSFULLY_RECONSTRUCTED_CODE :
            with lock :
                self.completely_processed_filepaths.append(dfc.filepath)
            if dfc.rootdir is not None :
                short_filepath = self.files_in_progress_by_path[dfc.filepath].full_filepath.relative_to(dfc.rootdir)
            else :
                short_filepath = self.files_in_progress_by_path[dfc.filepath].filepath
            self.logger.info(f'Getting message to produce for {short_filepath}...')
            new_msg = self._get_processing_result_message_for_file(self.files_in_progress_by_path[dfc.filepath],lock)
            if new_msg is not None :
                self.producer_message_queue.put(new_msg)
                return True
            self._failed_computing_processing_result(self.files_in_progress_by_path[dfc.filepath],self.lock)
            return False
        #otherwise the file is just in progress
        return True

    @abstractmethod
    def _get_processing_result_message_for_file(self,datafile,lock) :
        """
        Given a relative :class:`~.data_file_io.DownloadDataFileToMemory`, compute and return a
        :class:`~.data_file_io.ReproducerMessage` object that should be produced as the processing
        result for the file.

        This function should log an error and return None if the processing result fails to be computed.

        Not implemented in the base class.

        :param datafile: A :class:`~.data_file_io.DownloadDataFileToMemory` object that has received
            all of its messages from the topic
        :type datafile: :class:`~.data_file_io.DownloadDataFileToMemory`
        :param lock: Acquiring this :class:`threading.Lock` object would ensure that only one instance
            of :func:`~_get_processing_result_message_for_file` is running at once
        :type lock: :class:`threading.Lock`

        :return: message object to be produced
            (or None if computing it failed for any reason)
        :rtype: :class:`~.kafka_wrapper.Producible`
        """
        raise NotImplementedError

    def _failed_computing_processing_result(self,datafile,lock) :
        """
        This function is called when :func:`_get_processing_result_message_for_file` returns None because a
        processing result message could not be computed. It registers the original data file read from the
        topic as 'failed' in the .csv file so that that file will have its messages re-consumed if the
        program is restarted reading from the same topic with the same Consumer group ID. It then logs a
        warning and stops tracking the file.

        :param datafile: The datafile that should have been used to compute a processing result message
        :type datafile: :class:`~.data_file_io.DownloadDataFileToMemory`
        :param lock: Acquiring this :class:`threading.Lock` object would ensure that only one instance
            of :func:`~_get_processing_result_message_for_file` is running at once
        :type lock: :class:`threading.Lock`
        """
        #register the file in the registry as computing result failed
        with lock :
            self.file_registry.register_file_computing_result_failed(datafile.filename,
                                                                      datafile.filepath.relative_to(self._output_dir),
                                                                      datafile.n_total_chunks)
        #log the warning
        warnmsg = f'WARNING: Failed to compute the message to produce from {datafile.filepath}. Messages for this file '
        warnmsg+= 'will need to be re-consumed to try again. Check logs above for specific error messages.'
        self.logger.warning(warnmsg)
        #stop tracking the file
        with self.lock :
            del self.files_in_progress_by_path[datafile.filepath]
            del self.locks_by_fp[datafile.filepath]

    def _on_check(self) :
        msg = f'{self.n_msgs_read} messages read, {self.n_msgs_processed} messages processed, '
        msg+= f'{len(self.completely_processed_filepaths)} files completely reconstructed, and '
        msg+= f'{len(self.results_produced_filepaths)} processing result messages produced so far'
        self.logger.debug(msg)
        if ( len(self.files_in_progress_by_path)>0 or
             len(self.completely_processed_filepaths)>0 or
             len(self.results_produced_filepaths)>0 ) :
            self.logger.debug(self.progress_msg)

    @classmethod
    def get_command_line_arguments(cls):
        superargs,superkwargs = super().get_command_line_arguments()
        args = [*superargs,'consumer_topic_name','producer_topic_name','n_consumer_threads','n_producer_threads']
        kwargs = {**superkwargs}
        return args, kwargs



"""Example of a general Runnable class that can be installed as a Service/daemon"""

#imports
import datetime, time
from openmsistream.workflow import Runnable

class RunnableExample(Runnable) :
    """
    A class that extends Runnable as an example of installing a Runnable's
    "run_from_command_line" method as a Service/daemon
    """

    @classmethod
    def get_command_line_arguments(cls) :
        return ['output_dir'], {}

    @classmethod
    def run_from_command_line(cls,args=None) :
        parser = cls.get_argument_parser()
        args = parser.parse_args(args)
        test_file_name = 'runnable_example_service_test.txt'
        if not (args.output_dir/test_file_name).is_file() :
            with open(args.output_dir/test_file_name,'w') as fp :
                msg = "This file was created to test running a generic Runnable's 'run_from_command_line' "
                msg+= "method as a Service/daemon"
                fp.write(msg)
        with open(args.output_dir/test_file_name,'a') as fp :
            fp.write(f'\nService rerun {(datetime.datetime.now()).strftime("on %Y-%m-%d at %H:%M:%S")}')
        time.sleep(30)



"""Anything that has a classmethod to create an argument parser"""

#imports
from abc import ABC, abstractmethod

class HasArgumentParser(ABC) :
    """
    Base class for objects that have argument parsers
    """

    @classmethod
    @abstractmethod
    def get_argument_parser(cls) :
        """
        Return the argument parser that objects of this class use.

        Not implemented in the base class
        """
        raise NotImplementedError



"""Utility class holding constants and small calculations for working with Services/daemons"""

#imports
import pathlib, importlib
from inspect import isclass
import pkg_resources
from ..utilities import Logger

class ServicesConstants :
    """
    Constants for working with services
    """

    WORKING_DIR = (pathlib.Path(__file__).parent/'working_dir').resolve()
    NSSM_PATH = WORKING_DIR / 'nssm.exe'
    NSSM_DOWNLOAD_URL = 'https://nssm.cc/release/nssm-2.24.zip' # The URL to use for downloading NSSM when needed
    ERROR_LOG_STEM = '_ERROR_LOG.txt'
    SERVICE_EXECUTABLE_NAME_STEM = '_service_executable.py'
    DAEMON_SERVICE_DIR = pathlib.Path('/etc/systemd/system/')

    @property
    def available_services(self) :
        """
        A dictionary with details of the services that are available
        """
        return self.service_dicts

    def __init__(self) :
        #make the Service dictionaries to use
        self.service_dicts = []
        for script in pkg_resources.iter_entry_points('console_scripts') :
            if script.dist.key == 'openmsistream' :
                if script.name in ('InstallService','ManageService','ProvisionNode') :
                    continue
                scriptstr = str(script)
                cmd = (scriptstr.split())[0]
                path = ((scriptstr.split())[2].split(':'))[0]
                funcname = (((scriptstr.split())[2]).split(':'))[1]
                module = importlib.import_module(path)
                run_classes = [getattr(module,x) for x in dir(module)
                               if isclass(getattr(module,x)) and getattr(module,x).__name__==script.name]
                if len(run_classes)!=1 :
                    errmsg = f'ERROR: could not determine class for script {cmd} in file {path}! '
                    errmsg+= f'Possibilities found: {run_classes}'
                    raise RuntimeError(errmsg)
                self.service_dicts.append({'class_name':cmd,
                                           'class':run_classes[0],
                                           'filepath':path,
                                           'func_name':funcname})
        #make the logger to use
        self.logger = Logger('Services',logger_filepath=self.WORKING_DIR/'Services.log')

SERVICE_CONST = ServicesConstants()



"""Script to use the OS-appropriate ServceManager to install code as a Windows Service or Linux daemon"""

#imports
import sys
from .config import SERVICE_CONST
from .utilities import test_python_code, get_os_name
from .service_manager_base import ServiceManagerBase
from .windows_service_manager import WindowsServiceManager
from .linux_service_manager import LinuxServiceManager

def main(given_args=None) :
    """
    Install a Windows Service/Linux daemon (depending on the OS that's running)
    """
    #get the arguments
    if given_args is None :
        parser = ServiceManagerBase.get_argument_parser('install',sys.argv[1] if len(sys.argv)>1 else None)
        args, _ = parser.parse_known_args()
    else :
        parser = ServiceManagerBase.get_argument_parser('install',given_args[0] if len(given_args)>0 else None)
        args, _ = parser.parse_known_args(given_args)
    #run the tests if requested
    if args.service_spec_string=='test' :
        test_python_code()
    else :
        if args.service_name is None :
            #default name of the Service is just the class name
            service_name = args.service_spec_string
        else :
            service_name = args.service_name
        #make the list of arguments that should be sent to the run function in the executable (removing "service_name")
        argslist = sys.argv[2:] if given_args is None else given_args[1:]
        if '--service_name' in argslist :
            index = argslist.index('--service_name')
            argslist.pop(index)
            argslist.pop(index)
        #define arguments to and create the ServiceManager
        manager_args = [service_name]
        manager_kwargs = {'service_spec_string':args.service_spec_string,
                          'argslist':argslist,
                          'interactive':given_args is None,
                          'logger':SERVICE_CONST.logger}
        managers_by_os_name = {'Windows':WindowsServiceManager,
                               'Linux':LinuxServiceManager,}
        manager_class = managers_by_os_name[get_os_name()]
        manager = manager_class(*manager_args,**manager_kwargs)
        #install the service
        manager.install_service()

if __name__=='__main__' :
    main()



"""Example of how to extend a MetadataJSONReproducer for a real-world use case"""

#imports
from datetime import datetime
from ..metadata_json_reproducer import MetadataJSONReproducer

class XRDCSVMetadataReproducer(MetadataJSONReproducer) :
    """
    An example class showing how to use a MetadataJSONReproducer to extract metadata from the header
    of a .csv data file from an XRD measurement (read as chunks from a topic) and produce that metadata
    as JSON to another topic
    """

    def _get_metadata_dict_for_file(self,datafile) :
        #get the string of the file from the bytestring
        data_as_str = datafile.bytestring.decode()
        #split it into lines
        lines = data_as_str.split('\n')
        #read until the "[Measurement conditions]" block
        iline=0
        line = lines[iline].strip()
        while line!='[Measurement conditions]' :
            iline+=1
            if iline>=len(lines) or line=='[Scan points]':
                errmsg = f'ERROR: could not find a [Measurement Conditions] block in the "{datafile.full_filepath}" '
                errmsg+=  'file from which to extract metadata!'
                raise RuntimeError(errmsg)
            line = lines[iline].strip()
        #read lines up to the "[Scan points]" block and add each key/value to the dictionary
        iline+=1
        line=lines[iline].strip()
        metadata_dict = {}
        while line!='[Scan points]' and iline<len(lines) :
            linesplit = line.split(',')
            if len(linesplit)==2 :
                k,v = linesplit
                if v=='' :
                    metadata_dict[k]=None
                else :
                    metadata_dict[k]=v
            elif len(linesplit)>2 :
                k = linesplit[0]
                vlist = []
                for v in linesplit[1:] :
                    if v=='' :
                        vlist.append(None)
                    else :
                        vlist.append(v)
                metadata_dict[k] = vlist
            iline+=1
            line=lines[iline].strip()
        if not metadata_dict :
            warnmsg = f'WARNING: [Measurement Conditions] block in the "{datafile.full_filepath}" file did not contain '
            warnmsg+=  'any metadata keys/values!'
            self.logger.warning(warnmsg)
        #add a timestamp
        metadata_dict['metadata_message_generated_at'] = datetime.now().strftime('%m/%d/%Y, %H:%M:%S')
        #return the dictionary of metadata
        return metadata_dict

def main(args=None) :
    """
    main function to run the extended class from the command line
    """
    XRDCSVMetadataReproducer.run_from_command_line(args)

if __name__=='__main__' :
    main()



"""
The OpenMSIStream wrapper around standard Kafka in confluent_kafka.
Also includes wrapping around needed KafkaCrypto objects.
"""

from .openmsistream_producer import OpenMSIStreamProducer
from .openmsistream_consumer import OpenMSIStreamConsumer
from .producer_group import ProducerGroup
from .consumer_group import ConsumerGroup
from .consumer_and_producer_group import ConsumerAndProducerGroup
from .openmsistream_kafka_crypto import OpenMSIStreamKafkaCrypto
from .producible import Producible

__all__ = [
    'OpenMSIStreamProducer',
    'OpenMSIStreamConsumer',
    'ProducerGroup',
    'ConsumerGroup',
    'ConsumerAndProducerGroup',
    'OpenMSIStreamKafkaCrypto',
    'Producible',
]



"""
A process that will run while waiting for user input to check progress/status or shut it down
"""

#imports
import datetime
from queue import Queue, Empty
from threading import Thread
from abc import ABC, abstractmethod
from ..utilities.misc import add_user_input
from ..utilities.logging import LogOwner
from .config import RUN_CONST

class ControlledProcess(LogOwner,ABC) :
    """
    A class to use when running processes that should remain active until they are explicitly shut down

    :param update_secs: number of seconds to wait between printing a progress character to the console
        to indicate the program is alive
    :type update_secs: int, optional
    """

    #################### PROPERTIES ####################

    @property
    def alive(self) :
        """
        Read-only boolean indicating if the process is running
        """
        return self.__alive

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,*args,update_secs=RUN_CONST.DEFAULT_UPDATE_SECONDS,**other_kwargs) :
        self.__update_secs = update_secs
        #start up a Queue that will hold the control commands
        self.control_command_queue = Queue()
        #use a daemon thread to allow a user to input control commands from the command line
        #while the process is running
        user_input_thread = Thread(target=add_user_input,args=(self.control_command_queue,))
        user_input_thread.daemon=True
        user_input_thread.start()
        #a variable to indicate if the process has been shut down yet
        self.__alive = False
        #the last time the "still alive" character was printed
        self.__last_update = datetime.datetime.now()
        super().__init__(*args,**other_kwargs)

    def shutdown(self) :
        """
        Stop the process running.
        """
        self.control_command_queue.task_done()
        self.__alive = False
        self._on_shutdown()

    #################### PRIVATE HELPER FUNCTIONS ####################

    def _print_still_alive(self) :
        #print the "still alive" character
        if self.__update_secs!=-1 and (datetime.datetime.now()-self.__last_update).total_seconds()>self.__update_secs:
            self.logger.debug('.')
            self.__last_update = datetime.datetime.now()

    def _check_control_command_queue(self) :
        #if anything exists in the control command queue
        try :
            cmd = self.control_command_queue.get(block=True,timeout=0.5)
        except Empty :
            cmd = None
        if cmd is not None :
            if cmd.lower() in ('q','quit') : # shut down the process
                self.shutdown()
            elif cmd.lower() in ('c','check') : # run the on_check function
                self._on_check()
            else : # otherwise just skip this unrecognized command
                self._check_control_command_queue()

    #################### ABSTRACT METHODS ####################

    @abstractmethod
    def run(self) :
        """
        Classes extending this base class should include the logic of actually
        running the controlled process in this function, and should call super().run()
        before anything else to set some internal variables
        """
        self.__alive = True
        self.__last_update = datetime.datetime.now()

    @abstractmethod
    def _on_check(self) :
        """
        This function is run when the "check" command is found in the control queue.

        Not implemented in the base class
        """
        raise NotImplementedError

    @abstractmethod
    def _on_shutdown(self) :
        """
        This function is run when the process is stopped; it's called from :func:`~shutdown`.

        Not implemented in the base class
        """
        raise NotImplementedError



"""A set of Consumers sharing a subscribed topic, group ID, and (optionally) KafkaCrypto instance for key passing"""

#imports
import kafka
from confluent_kafka.admin import AdminClient
from confluent_kafka import TopicPartition
from ..utilities import LogOwner
from .utilities import reset_to_beginning_on_assign
from .config_file_parser import KafkaConfigFileParser
from .openmsistream_consumer import OpenMSIStreamConsumer

class ConsumerGroup(LogOwner) :
    """
    Class for working with a group of consumers sharing a single :class:`kafkacrypto.KafkaCrypto` instance

    :param config_path: Path to the config file that should be used to define Consumers in the group
    :type config_path: :class:`pathlib.Path`
    :param topic_name: The name of the topic to which the Consumers should be subscribed
    :type topic_name: str
    :param consumer_group_id: The ID string that should be used for each Consumer in the group.
        "create_new" (the default) will create a new UID to use.
    :type consumer_group_id: str, optional
    :param kafkacrypto: The :class:`~OpenMSIStreamKafkaCrypto` object that should be used to instantiate Consumers.
        Only needed if a single specific :class:`~OpenMSIStreamKafkaCrypto` instance should be shared.
    :type kafkacrypto: :class:`~OpenMSIStreamKafkaCrypto`, optional
    :param kwargs: Other keyword arguments will be added to the underlying Consumer's configurations,
        with underscores in their names replaced with dots.
    :type kwargs: dict
    """

    @property
    def topic_name(self) :
        """
        Name of the topic to which Consumers are subscribed
        """
        return self.__topic_name
    @property
    def consumer_group_id(self) :
        """
        String ID of Consumers in the group
        """
        return self.__consumer_group_id
    @property
    def kafkacrypto(self) :
        """
        The KafkaCrypto object handling key passing and deserialization for the group of Consumers (if applicable)
        """
        return self.__c_kwargs['kafkacrypto'] if 'kafkacrypto' in self.__c_kwargs else None

    def __init__(self,config_path,topic_name,*,consumer_group_id='create_new',kafkacrypto=None,**kwargs) :
        """
        Constructor method
        """
        super().__init__(**kwargs)
        self.__group_starting_offsets = self.__get_group_starting_offsets(config_path,topic_name,consumer_group_id)
        self.__topic_name = topic_name
        self.__c_args, self.__c_kwargs = OpenMSIStreamConsumer.get_consumer_args_kwargs(config_path,
                                                                                        group_id=consumer_group_id,
                                                                                        logger=self.logger,
                                                                                        kafkacrypto=kafkacrypto)
        if len(self.__c_args)>1 and 'group.id' in self.__c_args[1] :
            self.__consumer_group_id = self.__c_args[1]['group.id']
        else :
            self.__consumer_group_id = consumer_group_id

    def get_new_subscribed_consumer(self,*,restart_at_beginning=False,**kwargs) :
        """
        Return a new Consumer, subscribed to the topic and with the shared group ID.
        Call this function from a child thread to get thread-independent Consumers.

        Note: This function just creates and subscribes the Consumer. Polling it, closing
        it, and everything else must be handled by whatever calls this function.

        :param restart_at_beginning: if True, the new Consumer will start reading partitions from the earliest
            messages available, regardless of Consumer group ID and auto.offset.reset values.
            Useful when re-reading messages.
        :type restart_at_beginning: bool, optional
        :param kwargs: other keyword arguments are passed to the :class:`~OpenMSIStreamConsumer` constructor method
            (for example, parameters to set a key regex or how to filter messages)
        :type kwargs: dict

        :return: a Consumer created using the configs set in the constructor/from `kwargs`, subscribed to the topic
        :rtype: :class:`~OpenMSIStreamConsumer`
        """
        consumer = OpenMSIStreamConsumer(*self.__c_args,
                                         **kwargs,
                                         starting_offsets=self.__group_starting_offsets,**self.__c_kwargs)
        if restart_at_beginning :
            consumer.subscribe([self.__topic_name],on_assign=reset_to_beginning_on_assign)
        else :
            consumer.subscribe([self.__topic_name])
        return consumer

    def close(self) :
        """
        Wrapper around :func:`kafkacrypto.KafkaCrypto.close`.
        """
        try :
            self.__c_kwargs['kafkacrypto'].close()
        except Exception :
            pass
        finally :
            self.__c_kwargs['kafkacrypto'] = None

    def __get_group_starting_offsets(self,config_path,topic_name,consumer_group_id) :
        """
        Return a list of TopicPartitions listing the starting offsets for each partition
        in the topic for the given consumer group ID

        Re-raises any errors encountered in getting the necessary metadata,
        returning None if that happens
        """
        cfp = KafkaConfigFileParser(config_path)
        starting_offsets = []
        try :
            cluster_metadata = AdminClient(cfp.broker_configs).list_topics(topic=topic_name)
            n_partitions = len(cluster_metadata.topics[topic_name].partitions)
            if n_partitions<=0 :
                raise RuntimeError(f'ERROR: number of partitions for topic {topic_name} is {n_partitions}')
            kac_kwargs = {}
            for k,v in cfp.broker_configs.items() :
                if k in ('sasl.username','sasl.password') :
                    key = k.replace('.','_plain_')
                else :
                    key = k.replace('.','_')
                kac_kwargs[key]=v
            parts = [kafka.TopicPartition(topic_name,p_i) for p_i in range(n_partitions)]
            tp_offsets=kafka.KafkaAdminClient(**kac_kwargs).list_consumer_group_offsets(group_id=consumer_group_id,
                                                                                        partitions=parts)
            if len(tp_offsets)!=n_partitions :
                errmsg = f'Found {n_partitions} partitions for topic {topic_name} but got {len(tp_offsets)} '
                errmsg+= 'TopicPartitions listing current consumer group offsets'
                raise RuntimeError(errmsg)
            for t_p,offset_metadata in tp_offsets.items() :
                starting_offsets.append(TopicPartition(t_p.topic,t_p.partition,offset_metadata.offset))
            return starting_offsets
        except Exception as exc :
            errmsg = f'ERROR: encountered an exception when gathering initial "{topic_name}" topic offsets for '
            errmsg+= f'consumer group ID "{consumer_group_id}". The error will be logged below and re-raised.'
            self.logger.error(errmsg,exc_obj=exc)
            return None



"""
A wrapper around ConfigFileParser to handle OpenMSIStream config files that have sections
for connecting to brokers and running Producers and Consumers
"""

#imports
import pathlib
from confluent_kafka.serialization import DoubleSerializer, IntegerSerializer, StringSerializer
from confluent_kafka.serialization import DoubleDeserializer, IntegerDeserializer, StringDeserializer
from ..utilities.config_file_parser import ConfigFileParser
from ..workflow.config import RUN_CONST
from .serialization import DataFileChunkSerializer, DataFileChunkDeserializer

class KafkaConfigFileParser(ConfigFileParser) :
    """
    A class to parse Kafka configurations from files
    """

    #################### CONSTANTS ####################

    SERIALIZERS = [
        StringSerializer,
        IntegerSerializer,
        DoubleSerializer,
        DataFileChunkSerializer,
    ]

    DESERIALIZERS = [
        StringDeserializer,
        IntegerDeserializer,
        DoubleDeserializer,
        DataFileChunkDeserializer,
    ]

    #################### PROPERTIES ####################

    @property
    def broker_configs(self) :
        """
        Configs for connecting to the broker
        """
        if self.__broker_configs is None :
            self.__broker_configs = self._get_config_dict('broker')
        return self.__convert_floats(self.__broker_configs)
    @property
    def producer_configs(self) :
        """
        Configs for setting up Producers
        """
        if self.__producer_configs is None :
            pcs = self._get_config_dict('producer')
            self.__producer_configs = self.get_replaced_configs(pcs,'serialization')
        return self.__convert_floats(self.__producer_configs)
    @property
    def consumer_configs(self) :
        """
        Configs for setting up Consumers
        """
        if self.__consumer_configs is None :
            ccs = self._get_config_dict('consumer')
            #if the auto.offset.reset was given as "none" then remove it from the ccs
            if 'auto.offset.reset' in ccs and ccs['auto.offset.reset']=='none' :
                del ccs['auto.offset.reset']
            self.__consumer_configs = self.get_replaced_configs(ccs,'deserialization')
        return self.__convert_floats(self.__consumer_configs)

    #################### PUBLIC FUNCTIONS ####################

    def __init__(self,*args,**kwargs) :
        super().__init__(*args,**kwargs)
        self.__broker_configs = None
        self.__producer_configs = None
        self.__consumer_configs = None
        self.kc_config_file_str = self.__get_kc_config_file_str()

    @staticmethod
    def get_replaced_configs(configs,replacement_type) :
        """
        Returns a configuration dictionary with (de)serialization parameters replaced
        by instances of corresponding classes

        configs = the configurations dictionary to alter and return
        replacement_type = a string indicating the type of replacement that should be performed
        """
        if replacement_type=='serialization' :
            classes = KafkaConfigFileParser.SERIALIZERS
        elif replacement_type=='deserialization' :
            classes = KafkaConfigFileParser.DESERIALIZERS
        else :
            raise ValueError(f'ERROR: unrecognized replacement_type "{replacement_type}" in get_replaced_configs!')
        for cfg_name,cfg_value in configs.items() :
            for serdes_class in classes :
                if cfg_value==serdes_class.__name__ :
                    configs[cfg_name] = serdes_class()
                    break
        return configs

    #################### PRIVATE HELPER FUNCTIONS ####################

    def __get_kc_config_file_str(self) :
        """
        Returns the path to the config file that KafkaCrypto needs based on the current configs.
        Return value is a string as expected by KafkaCrypto.
        If no config file is found according to the conventions listed below, this function returns None
        and it will be assumed that no configuration for KafkaCrypto exists

        Options are:
        1) The regular config file has a "kafkacrypto" section with a "config_file" parameter that is the
        path to the KafkaCrypo config file
        2) The regular config file has a "kafkacrypto" section with a "node_id" parameter corresponding to
        a named subdirectory in openmsistream/kafka_wrapper/config_files that was created when the node was provisioned
        """
        if 'kafkacrypto' in self.available_group_names :
            kc_configs = self.get_config_dict_for_groups('kafkacrypto')
            #option 1 above
            if 'config_file' in kc_configs :
                path_as_str = (kc_configs['config_file']).lstrip('file#')
                if not pathlib.Path(path_as_str).is_file() :
                    errmsg = f'ERROR: KafkaCrypto config file {path_as_str} '
                    errmsg+= f'(from config file {self.filepath}) not found!'
                    self.logger.error(errmsg,FileNotFoundError)
                return path_as_str
            #option 2 above
            if 'node_id' in kc_configs :
                node_id = kc_configs['node_id']
                dirpath = RUN_CONST.CONFIG_FILE_DIR / node_id
                filepath = dirpath / f'{node_id}.config'
                if (not dirpath.is_dir()) or (not filepath.is_file()) :
                    errmsg = f'ERROR: no KafkaCrypto config file found in the default location ({filepath}) '
                    errmsg+= f'for node ID = {node_id}'
                    self.logger.error(errmsg,FileNotFoundError)
                return str(filepath)
        #no config file found
        return None

    def __convert_floats(self,given_dict) :
        """
        Return a version of the dictionary "given_dict" where any values that can be converted to floats
        are converted to floats
        (Some KafkaCrypto code assumes configs will be floats)
        """
        return_dict = {}
        for k,v in given_dict.items() :
            try :
                v_as_float = float(v)
                return_dict[k] = v_as_float
            except Exception :
                return_dict[k]=v
        return return_dict



"""The base message type that must be produced by a MetadataJSONReproducer"""

#imports
import json
from ..data_file_io.entity.reproducer_message import ReproducerMessage

class MetadataJSONMessage(ReproducerMessage) :
    """
    A class to represent the json metadata from a file as a Producible
    """

    @property
    def msg_value(self) :
        if self.json_content is None :
            raise ValueError(f'ERROR: JSON content should not be None for a {self.__class__.__name__}!')
        return json.dumps(self.json_content)

    def __init__(self,*args,json_content=None,**kwargs) :
        super().__init__(*args,**kwargs)
        self.json_content=json_content



"""Base class for consuming file chunks into memory and then triggering some action when whole files are available"""

#imports
import pathlib, warnings
from abc import ABC
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from kafkacrypto.message import KafkaCryptoMessage
from ...utilities.misc import populated_kwargs
from ...workflow import Runnable
from ..config import DATA_FILE_HANDLING_CONST
from ..utilities import get_encrypted_message_timestamp_string
from ..entity.download_data_file import DownloadDataFileToMemory
from .data_file_chunk_handlers import DataFileChunkHandler

class DataFileStreamHandler(DataFileChunkHandler,Runnable,ABC) :
    """
    Generic base class for consuming file chunks into memory and then doing something once entire files are available
    """

    def __init__(self,*args,output_dir=None,datafile_type=DownloadDataFileToMemory,**kwargs) :
        """
        Constructor method
        """
        #make sure the directory for the output is set
        self._output_dir = self._get_auto_output_dir() if output_dir is None else output_dir
        if not self._output_dir.is_dir() :
            self._output_dir.mkdir(parents=True)
        kwargs = populated_kwargs(kwargs,{'logger_file':self._output_dir})
        super().__init__(*args,datafile_type=datafile_type,**kwargs)
        self.logger.info(f'Log files and output will be in {self._output_dir}')
        if not issubclass(datafile_type,DownloadDataFileToMemory) :
            errmsg = f'ERROR: {self.__class__.__name__} requires a datafile_type that is a subclass of '
            errmsg+= f'DownloadDataFileToMemory but {datafile_type} was given!'
            self.logger.error(errmsg,ValueError)
        self.file_registry = None # needs to be set in subclasses

    def _process_message(self,lock,msg,rootdir_to_set=None):
        """
        Parent class message processing function to check for:
            undecryptable messages (returns False)
            files where reconstruction is just in progress (returns True)
            files with mismatched hashes (returns False)
        Child classes should call super()._process_message() and check the return value
        to find successfully-reconstructed files for further handling.
        """
        retval = super()._process_message(lock, msg, self._output_dir if rootdir_to_set is None else rootdir_to_set)
        #if the message was returned because it couldn't be decrypted, write it to the encrypted messages directory
        if ( hasattr(retval,'key') and hasattr(retval,'value') and
             (isinstance(retval.key,KafkaCryptoMessage) or isinstance(retval.value,KafkaCryptoMessage)) ) :
            self._undecryptable_message_callback(retval)
            return False
        #get the DataFileChunk from the message value
        try :
            dfc = msg.value() #from a regular Kafka Consumer
        except TypeError :
            dfc = msg.value #from KafkaCrypto
        #if the file is just in progress
        if retval is True :
            with lock :
                self.file_registry.register_file_in_progress(dfc)
            return retval
        #if the file hashes didn't match, invoke the callback and return False
        if retval==DATA_FILE_HANDLING_CONST.FILE_HASH_MISMATCH_CODE :
            errmsg = f'ERROR: hashes for file {self.files_in_progress_by_path[dfc.filepath].filename} not matched '
            errmsg+= 'after being fully read! The messages for this file will need to be consumed again if the file '
            errmsg+= 'is to be processed! Please rerun with the same consumer ID to try again.'
            self.logger.error(errmsg)
            with lock :
                self.file_registry.register_file_mismatched_hash(dfc)
            self._mismatched_hash_callback(self.files_in_progress_by_path[dfc.filepath],lock)
            with lock :
                del self.files_in_progress_by_path[dfc.filepath]
                del self.locks_by_fp[dfc.filepath]
            return False
        if retval!=DATA_FILE_HANDLING_CONST.FILE_SUCCESSFULLY_RECONSTRUCTED_CODE :
            self.logger.error(f'ERROR: unrecognized add_chunk return value: {retval}',NotImplementedError)
            return False
        return retval

    def _undecryptable_message_callback(self,msg) :
        """
        This function is called when a message that could not be decrypted is found.
        If this function is called it is likely that the file the chunk is coming from won't be able to be processed.

        In the base class, this logs a warning.

        :param msg: the :class:`kafkacrypto.Message` object with undecrypted :class:`kafkacrypto.KafkaCryptoMessages`
            for its key and/or value
        :type msg: :class:`kafkacrypto.Message`
        """
        timestamp_string = get_encrypted_message_timestamp_string(msg)
        warnmsg = f'WARNING: encountered a message that failed to be decrypted (timestamp = {timestamp_string}). '
        warnmsg+= 'This message will be skipped, and the file it came from cannot be processed from the stream '
        warnmsg+= 'until it is decryptable. Please rerun with a new Consumer ID to consume these messages again.'
        self.logger.warning(warnmsg)

    def _mismatched_hash_callback(self,datafile,lock) :
        """
        Called when a file reconstructed in memory doesn't match the hash of its contents originally on disk,
        providing an opportunity for fallback/backup processing in the case of failure.

        Does nothing in the base class.

        :param datafile: A :class:`~.data_file_io.DownloadDataFileToMemory` object that has received
            all of its messages from the topic
        :type datafile: :class:`~.data_file_io.DownloadDataFileToMemory`
        :param lock: Acquiring this :class:`threading.Lock` object would ensure that only one instance
            of :func:`~_mismatched_hash_callback` is running at once
        :type lock: :class:`threading.Lock`
        """

    @classmethod
    def _get_auto_output_dir(cls) :
        return pathlib.Path()/f'{cls.__name__}_output'

    @classmethod
    def get_command_line_arguments(cls):
        args = ['config','consumer_group_id','update_seconds']
        kwargs = {'optional_output_dir': cls._get_auto_output_dir(),}
        return args, kwargs



"""A wrapped Kafka Producer. May produce encrypted messages."""

#imports
import time, warnings
from confluent_kafka import SerializingProducer
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from kafkacrypto import KafkaProducer
from ..utilities import LogOwner
from ..utilities.misc import raise_err_with_optional_logger, debug_msg_with_optional_logger
from .utilities import add_kwargs_to_configs, default_producer_callback, make_callback
from .producible import Producible
from .config_file_parser import KafkaConfigFileParser
from .openmsistream_kafka_crypto import OpenMSIStreamKafkaCrypto
from .serialization import CompoundSerializer

class OpenMSIStreamProducer(LogOwner) :
    """
    Wrapper for working with a Producer of some type. Expects message values that are
    :class:`~.data_file_io.entity.data_file_chunk.DataFileChunk` objects by default;
    other message value types can be accommodated by setting "value.serializer" in the config file.

    :param producer_type: The type of underlying Producer that should be used
    :type producer_type: :class:`confluent_kafka.SerializingProducer` or :class:`kafkacrypto.KafkaProducer`
    :param configs: A dictionary of configuration names and parameters to use in instantiating the underlying Producer
    :type configs: dict
    :param kafkacrypto: The :class:`~.kafka_wrapper.OpenMSIStreamKafkaCrypto` object that should be used
        to instantiate the Producer. Only needed if `producer_type` is :class:`kafkacrypto.KafkaProducer`.
    :type kafkacrypto: :class:`~.kafka_wrapper.OpenMSIStreamKafkaCrypto`, optional
    :param kwargs: Any extra keyword arguments (other than "logger") are added to the configuration dict for the
        Producer, with underscores in their names replaced by dots
    :type kwargs: dict

    :raises ValueError: if `producer_type` is not :class:`confluent_kafka.SerializingProducer`
        or :class:`kafkacrypto.KafkaProducer`
    :raises ValueError: if `producer_type` is :class:`kafkacrypto.KafkaProducer` and `kafkacrypto` is None
    """

    POLL_EVERY = 5 # poll the producer at least every 5 calls to produce

    def __init__(self,producer_type,configs,kafkacrypto=None,**kwargs) :
        """
        Constructor method
        """
        super().__init__(**kwargs)
        if producer_type==KafkaProducer :
            if kafkacrypto is None :
                errmsg = 'ERROR: creating a KafkaProducer requires holding onto its KafkaCrypto objects!'
                self.logger.error(errmsg,ValueError)
            self.__kafkacrypto = kafkacrypto
            self._producer = producer_type(**configs)
        elif producer_type==SerializingProducer :
            self._producer = producer_type(configs)
        else :
            errmsg=f'ERROR: Unrecognized producer type {producer_type} for OpenMSIStreamProducer!'
            self.logger.error(errmsg,ValueError)
        # poll the producer at least every 5 calls to produce
        self.__poll_counter = 0

    @staticmethod
    def get_producer_args_kwargs(config_file_path,logger=None,kafkacrypto=None,**kwargs) :
        """
        Return the list of arguments and dictionary or keyword arguments that should be used to instantiate
        :class:`~OpenMSIStreamProducer` objects based on the given config file.
        Used to share a single :class:`~.kafka_wrapper.OpenMSIStreamKafkaCrypto` instance across
        several Producers.

        :param config_file_path: Path to the config file to use in defining Producers
        :type config_file_path: :class:`pathlib.Path`
        :param logger: The :class:`~.utilities.Logger` object to use for each of the
            :class:`~OpenMSIStreamProducer` objects
        :type logger: :class:`~.utilities.Logger`
        :param kafkacrypto: The :class:`~.kafka_wrapper.OpenMSIStreamKafkaCrypto` object that should
            be used to instantiate Producers. Only needed if a single specific
            :class:`~.kafka_wrapper.OpenMSIStreamKafkaCrypto` instance should be shared.
        :type kafkacrypto: :class:`~.kafka_wrapper.OpenMSIStreamKafkaCrypto`, optional
        :param kwargs: Any extra keyword arguments are added to the configuration dict for the Producers,
            with underscores in their names replaced by dots
        :type kwargs: dict

        :return: ret_args, the list of arguments to create new :class:`~OpenMSIStreamProducer` objects
            based on the config file and arguments to this method
        :rtype: list
        :return: ret_kwargs, the dictionary of keyword arguments to create new :class:`~OpenMSIStreamProducer` objects
            based on the config file and arguments to this method
        :rtype: dict
        """
        parser = KafkaConfigFileParser(config_file_path,logger=logger)
        ret_kwargs = {'logger':logger}
        #get the broker and producer configurations
        all_configs = {**parser.broker_configs,**parser.producer_configs}
        all_configs = add_kwargs_to_configs(all_configs,logger,**kwargs)
        #all_configs['debug']='broker,topic,msg'
        #use a SerializingProducer by default
        if kafkacrypto is None and parser.kc_config_file_str is None :
            ret_args = [SerializingProducer,all_configs]
            return ret_args, ret_kwargs
        #if "kafkacrypto" is given, or there are configs for KafkaCrypto, use a KafkaProducer
        if kafkacrypto is not None :
            if not isinstance(kafkacrypto,OpenMSIStreamKafkaCrypto) :
                errmsg = f'ERROR: kafkacrypto argument type = {type(kafkacrypto)}'
                raise_err_with_optional_logger(logger,errmsg,TypeError)
            debugmsg = f'Produced messages will be encrypted using configs at {kafkacrypto.config_file_path}'
            debug_msg_with_optional_logger(logger,debugmsg)
            k_c = kafkacrypto
        else :
            debugmsg = f'Produced messages will be encrypted using configs at {parser.kc_config_file_str}'
            debug_msg_with_optional_logger(logger,debugmsg)
            k_c = OpenMSIStreamKafkaCrypto(parser.broker_configs,parser.kc_config_file_str)
        if 'key.serializer' in all_configs :
            keyser = CompoundSerializer(all_configs.pop('key.serializer'),k_c.key_serializer)
        else :
            keyser = k_c.key_serializer
        if 'value.serializer' in all_configs :
            valser = CompoundSerializer(all_configs.pop('value.serializer'),k_c.value_serializer)
        else :
            valser = k_c.value_serializer
        all_configs['key_serializer']=keyser
        all_configs['value_serializer']=valser
        ret_args = [KafkaProducer,all_configs]
        ret_kwargs['kafkacrypto']=k_c
        return ret_args, ret_kwargs

    @classmethod
    def from_file(cls,*args,**kwargs) :
        """
        Wrapper around :func:`~OpenMSIStreamProducer.get_producer_args_kwargs` and the :class:`~OpenMSIStreamProducer`
        constructor to return a single :class:`~OpenMSIStreamProducer` from a given config file/arguments.
        Arguments are the same as :func:`~OpenMSIStreamProducer.get_producer_args_kwargs`

        :returns: An :class:`~OpenMSIStreamProducer` object based on the given config file/arguments
        :rtype: :class:`~OpenMSIStreamProducer`
        """
        args_to_use, kwargs_to_use = OpenMSIStreamProducer.get_producer_args_kwargs(*args,**kwargs)
        return cls(*args_to_use,**kwargs_to_use)

    def produce_from_queue(self,queue,topic_name,**kwargs) :
        """
        Get a :class:`~.kafka_wrapper.Producible` object from a given Queue and produce it to
        the given topic. Does nothing if the queue is empty, and does not block waiting for items from the queue.

        Meant to be run in multiple threads in parallel.

        :param queue: the :class:`queue.Queue` holding objects that should be produced
        :type queue: :class:`queue.Queue`
        :param topic_name: the name of the topic to produce to
        :type topic_name: str
        :param kwargs: other keyword arguments are passed to :func:`~produce_object`
        :type kwargs: dict
        """
        if queue.empty() :
            return
        #get the next object from the Queue
        obj = queue.get_nowait()
        if isinstance(obj,Producible) :
            success = self.produce_object(obj,topic_name,**kwargs)
            if not success :
                warnmsg = f'WARNING: message with key {obj.msg_key} failed to buffer for longer than '
                warnmsg+=  'the timeout with no new callbacks served. This message will be re-enqueued.'
                self.logger.warning(warnmsg)
                queue.put(obj)
            self.__poll_counter+=1
            if self.__poll_counter%self.POLL_EVERY==0 :
                _ = self.poll(0)
                self.__poll_counter = 0
        else :
            warnmsg = f'WARNING: found an object of type {type(obj)} in a Producer queue that should only contain '
            warnmsg+= 'Producible objects. This object will be skipped!'
            self.logger.warning(warnmsg)

    def produce_from_queue_looped(self,queue,topic_name,**kwargs) :
        """
        Get :class:`~.kafka_wrapper.Producible` objects from a given Queue and produce
        them to the given topic. Blocks while waiting for items to appear in the queue.
        Runs until "None" is pulled from the queue.

        Meant to be run in multiple threads in parallel.

        :param queue: the :class:`queue.Queue` holding objects that should be produced
        :type queue: :class:`queue.Queue`
        :param topic_name: the name of the topic to produce to
        :type topic_name: str
        :param kwargs: other keyword arguments are passed to :func:`~produce_object`
        :type kwargs: dict
        """
        #get the next object from the Queue
        obj = queue.get()
        #loop until "None" is pulled from the Queue
        while obj is not None :
            if isinstance(obj,Producible) :
                success = self.produce_object(obj,topic_name,**kwargs)
                if not success :
                    warnmsg = f'WARNING: message with key {obj.msg_key} failed to buffer for longer than '
                    warnmsg+=  'the timeout with no new callbacks served. This message will be re-enqueued.'
                    self.logger.warning(warnmsg)
                    queue.put(obj)
                self.__poll_counter+=1
                if self.__poll_counter%self.POLL_EVERY==0 :
                    _ = self.poll(0)
                    self.__poll_counter = 0
            else :
                warnmsg = f'WARNING: found an object of type {type(obj)} in a Producer queue that should only contain '
                warnmsg+= 'Producible objects. This object will be skipped!'
                self.logger.warning(warnmsg)
            #get the next object in the Queue
            obj = queue.get()
        queue.task_done()

    def produce_object(self,obj,topic_name,callback=None,print_every=1000,timeout=60,retry_sleep=5) :
        """
        Produce a given :class:`~.kafka_wrapper.Producible` object to a given topic,
        with some handling for BufferErrors, calling poll() automatically, and using callbacks
        constructed on the fly.

        :param obj: the object to produce
        :type obj: :class:`~.kafka_wrapper.Producible`
        :param topic_name: the name of the topic to produce to
        :type topic_name: str
        :param callback: a function that should be called for each message upon recognition by the broker.
            Will be wrapped in a lambda for each call to produce().
            Arguments to the callback function are determined by the particular type of :class:`Producible` object used
        :type callback: producer callback function (takes at least "err" and "msg" arguments), optional
        :param print_every: print/log progress every (this many) messages
        :type print_every: int, optional
        :param timeout: max time (seconds) to wait for the message to be produced in the event of
            (repeated) BufferError(s)
        :type timeout: float, optional
        :param retry_sleep: how long (seconds) to wait between produce attempts if one fails with a BufferError
        :type retry_sleep: float, optional

        :return: True if the call to produce is successful, False otherwise.
        :rtype: bool
        """
        #log a line about this message if applicable
        logmsg = obj.get_log_msg(print_every)
        if logmsg is not None :
            self.logger.info(logmsg)
        #get the Producible's callback arguments and set the callback to use
        if callback is None :
            callback_to_use = make_callback(default_producer_callback,logger=self.logger,**obj.callback_kwargs)
        else :
            callback_to_use = make_callback(callback,**obj.callback_kwargs)
        #produce the message to the topic
        success=False
        total_wait_secs=0
        while (not success) and total_wait_secs<timeout :
            try :
                self.produce(topic=topic_name,key=obj.msg_key,value=obj.msg_value,on_delivery=callback_to_use)
                success=True
            except BufferError :
                n_new_callbacks = self.poll(0)
                time.sleep(retry_sleep)
                if n_new_callbacks is None or n_new_callbacks==0 :
                    total_wait_secs+=retry_sleep
                else :
                    total_wait_secs = 0
        self.__poll_counter+=1
        if self.__poll_counter%self.POLL_EVERY==0 :
            n_new_callbacks = self.poll(0)
            self.__poll_counter = 0
        return success

    def produce(self,topic,key,value,**kwargs) :
        """
        Produce a message to a topic.
        Other kwargs are passed through to the underlying producer's produce() function.

        :param topic: the name of the topic to produce to
        :type topic: str
        :param key: the key of the message
        :type key: depends on serialization settings
        :param value: the value of the message
        :type value: depends on the serialization settings
        """
        try :
            if isinstance(self._producer,KafkaProducer) :
                key = self._producer.ks(topic,key)
                value = self._producer.vs(topic,value)
            return self._producer.produce(topic=topic,key=key,value=value,**kwargs)
        except Exception as exc :
            self.logger.error('ERROR: failed during call to Producer.produce! Will log and raise Exception.',
                              exc_obj=exc)
        return None

    def poll(self,*args,**kwargs) :
        """
        Wrapper around Producer.poll()
        """
        return self._producer.poll(*args,**kwargs)

    def flush(self,*args,**kwargs) :
        """
        Wrapper around Producer.flush()
        """
        return self._producer.flush(*args,**kwargs)

    def close(self) :
        """
        Wrapper around :func:`kafkacrypto.KafkaCrypto.close`.
        It's important to call this on shutdown if the Producer is producing encrypted messages.
        """
        try :
            if self.__kafkacrypto :
                self.__kafkacrypto.close()
        except Exception :
            pass
        finally :
            self.__kafkacrypto = None



"""
A ConsumerAndProducerGroup whose receipt and production of messages
is managed using the ControlledProcessInfrastructure
"""

#imports
import time
from abc import ABC, abstractmethod
from queue import Queue
from ..workflow.config import RUN_CONST
from ..workflow.controlled_process_multi_threaded import ControlledProcessMultiThreaded
from .consumer_and_producer_group import ConsumerAndProducerGroup

class ControlledMessageReproducer(ControlledProcessMultiThreaded,ConsumerAndProducerGroup,ABC) :
    """
    An abstract base class combining and ControlledProcessMultiThreaded and a ConsumerAndProducerGroup
    to systematically read messages and produce others.
    """

    CONSUMER_POLL_TIMEOUT = 0.050
    NO_MESSAGE_WAIT = 0.005
    FLUSH_PRODUCER_EVERY = 100 #flush the producer after this many calls to produce_from_queue (could be fast)
    PRODUCER_FLUSH_TIMEOUT = 0.050 #timeout for the intermediate calls to producer.flush

    def __init__(self,config_path,consumer_topic_name,producer_topic_name,*,
                 n_producer_threads=1,n_consumer_threads=RUN_CONST.DEFAULT_N_THREADS,**kwargs) :
        """
        Hang onto the number of messages read, processed, and produced
        """
        self.n_msgs_read = 0
        self.n_msgs_processed = 0
        self.n_msgs_produced = 0
        super().__init__(config_path,consumer_topic_name,n_threads=max(n_producer_threads,n_consumer_threads),**kwargs)
        self.restart_at_beginning = False #set to true to reset new consumers to their earliest offsets
        self.message_key_regex = None #set to some regex to filter messages by their keys
        self.filter_new_messages = False #reset the regex after the consumer has filtered through previous messages
        self.last_message = None #hold onto the last consumed message to commit its offset on shutdown
        self.producer_topic_name = producer_topic_name
        self.n_producer_threads = n_producer_threads
        self.n_consumer_threads = n_consumer_threads
        self.producer_message_queue = Queue()

    def _run_worker(self,create_consumer=False,create_producer=False,
                    produce_from_queue_args=None,produce_from_queue_kwargs=None):
        """
        Handle optional startup and shutdown of thread-independent Consumer and/or Producer.
        Serve individual messages to the _process_message function on the Consumer side,
        and produce messages from the shared Queue on the Producer side.
        """
        #create the Consumer and/or Producer for this thread
        if self.alive :
            consumer, producer = self.__get_thread_consumer_and_producer(create_consumer,create_producer)
        #start the loop for while the controlled process is alive
        calls_since_producer_flush = 0
        while self.alive :
            #if this thread has a Consumer side
            if consumer is not None :
                self.__consume_messages_while_alive(consumer)
            #if this thread has a Producer side
            if producer is not None :
                calls_since_producer_flush = self.__produce_messages_while_alive(producer,
                                                                                 produce_from_queue_args,
                                                                                 produce_from_queue_kwargs,
                                                                                 calls_since_producer_flush)
        #commit the offset of the last message received if it wasn't already consumed in the past (block until done)
        if ( (consumer is not None) and
             (self.last_message is not None) and (not consumer.message_consumed_before(self.last_message)) ) :
            self.__commit_last_message_offset(consumer)
        #shut down the Consumer/Producer that was created once the process isn't alive anymore
        if consumer is not None :
            consumer.close()
        if producer is not None :
            producer.flush(timeout=-1)
            producer.close()

    def _on_shutdown(self):
        super()._on_shutdown()
        self.close()

    @abstractmethod
    def _process_message(self,lock,msg,*args,**kwargs) :
        """
        Process a single message read from the thread-independent Consumer
        Returns true if processing was successful, and False otherwise

        lock = lock across all created child threads (use to enforce thread safety during processing)
        msg  = a single message that was consumed and should be processed by this function

        Not implemented in the base class
        """
        raise NotImplementedError

    def __get_thread_consumer_and_producer(self,create_consumer,create_producer) :
        """
        Return the consumer and producer to use for a specific thread
        """
        if create_consumer :
            consumer = self.get_new_subscribed_consumer(restart_at_beginning=self.restart_at_beginning,
                                                        message_key_regex=self.message_key_regex,
                                                        filter_new_messages=self.filter_new_messages)
        else :
            consumer = None
        if create_producer :
            producer = self.get_new_producer()
        else :
            producer = None
        if ( (consumer is not None) and
             ( ('enable.auto.commit' not in consumer.configs.keys()) or
               (consumer.configs['enable.auto.commit'] is True))) :
            warnmsg = 'WARNING: enable.auto.commit has not been set to False for a Consumer that will manually commit '
            warnmsg+= 'offsets. Missed or duplicate messages could result. You can set "enable.auto.commit"=False in '
            warnmsg+= 'the "consumer" section of the config file to re-enable manual offset commits (recommended).'
            self.logger.warning(warnmsg)
        return consumer, producer

    def __consume_messages_while_alive(self,consumer) :
        """
        Runs continuously while the calling thread is alive to consume messages
        """
        #consume a message from the topic
        msg = consumer.get_next_message(self.CONSUMER_POLL_TIMEOUT)
        if msg is None :
            time.sleep(self.NO_MESSAGE_WAIT) #wait just a bit to not over-tax things
            return
        with self.lock :
            self.n_msgs_read+=1
            self.last_message = msg
        #send the message to the _process_message function
        retval = self._process_message(self.lock,msg)
        #count and (asynchronously) commit the message as processed (if not consumed already in the past)
        if retval :
            with self.lock :
                self.n_msgs_processed+=1
            if not consumer.message_consumed_before(msg) :
                tps = consumer.commit(msg)
                if tps is None :
                    return
                for t_p in tps :
                    if t_p.error is not None :
                        warnmsg = 'WARNING: failed to synchronously commit offset of last message '
                        warnmsg+= f'received on "{t_p.topic}" partition {t_p.partition}. Duplicate '
                        warnmsg+=  'messages may result when this Consumer is restarted. '
                        warnmsg+= f'Error reason: {t_p.error.str()}'
                        self.logger.warning(warnmsg)

    def __produce_messages_while_alive(self,producer,pfq_args,pfq_kwargs,calls_since_flush) :
        """
        Runs continuously while the calling thread is alive to produce messages
        """
        if self.producer_message_queue.empty() :
            time.sleep(self.NO_MESSAGE_WAIT) #wait just a bit to not overtax things
        else :
            pfq_args = [] if pfq_args is None else pfq_args
            pfq_kwargs = {} if pfq_kwargs is None else pfq_kwargs
            producer.produce_from_queue(self.producer_message_queue,self.producer_topic_name,
                                        *pfq_args,**pfq_kwargs)
        if calls_since_flush>=self.FLUSH_PRODUCER_EVERY :
            producer.flush(timeout=self.PRODUCER_FLUSH_TIMEOUT)
            calls_since_flush=0
        else :
            calls_since_flush+=1
        return calls_since_flush

    def __commit_last_message_offset(self,consumer) :
        """
        Commit the offset of the last message received while blocking
        """
        try :
            tps = consumer.commit(self.last_message,asynchronous=False)
            if tps is None :
                return
            for t_p in tps :
                if t_p.error is not None :
                    warnmsg = 'WARNING: failed to synchronously commit offset of last message received on '
                    warnmsg+= f'"{t_p.topic}" partition {t_p.partition}. Duplicate messages may result '
                    warnmsg+= f'when this Consumer is restarted. Error reason: {t_p.error.str()}'
                    self.logger.warning(warnmsg)
        except Exception as exc :
            errmsg = 'WARNING: failed to synchronously commit offset of last message received. '
            errmsg+= 'Duplicate messages may be read the next time this Consumer is started. '
            errmsg+= 'Error will be logged below but not re-raised.'
            self.logger.error(errmsg,exc_obj=exc,reraise=False)



"""Expands a ConfigFileParser for the S3 section of configs"""

#imports
from ..utilities.config_file_parser import ConfigFileParser

class S3ConfigFileParser(ConfigFileParser) :
    """
    A class to parse S3 bucket configurations from files
    """

    @property
    def s3_configs(self) :
        """
        The S3 section of the config file
        """
        if self.__s3_configs is None :
            self.__s3_configs = self._get_config_dict('s3')
        return self.__s3_configs

    def __init__(self,*args,**kwargs) :
        super().__init__(*args,**kwargs)
        self.__s3_configs = None



"""
Base class for extracting metadata dictionaries from files in topics
and producing those dictionaries as JSON-formatted strings to another topic
"""

#imports
from abc import ABC, abstractmethod
from ..data_file_io.actor.data_file_stream_reproducer import DataFileStreamReproducer
from .metadata_json_message import MetadataJSONMessage

class MetadataJSONReproducer(DataFileStreamReproducer,ABC) :
    """
    A class to extend for processing a stream of DataFileChunks, extracting metadata from
    fully-reconstructed files, and producing those metadata fields as JSON to a different topic.

    This is a base class that cannot be instantiated on its own.

    :param config_path: Path to the config file to use in defining the Broker connection, Consumers, and Producers
    :type config_path: :class:`pathlib.Path`
    :param consumer_topic_name: Name of the topic to which the Consumer(s) should be subscribed
    :type consumer_topic_name: str
    :param producer_topic_name: Name of the topic to which the Producer(s) should produce the processing results
    :type producer_topic_name: str
    :param output_dir: Path to the directory where the log and csv registry files should be kept (if None a default
        will be created in the current directory)
    :type output_dir: :class:`pathlib.Path`, optional
    :param datafile_type: the type of data file that recognized files should be reconstructed as
        (must be a subclass of :class:`~.data_file_io.DownloadDataFileToMemory`)
    :type datafile_type: :class:`~.data_file_io.DownloadDataFileToMemory`, optional
    :param n_producer_threads: the number of producers to run. The total number of producer/consumer threads
        started is `max(n_consumer_threads,n_producer_threads)`.
    :type n_producer_threads: int, optional
    :param n_consumer_threads: the number of consumers to run. The total number of producer/consumer threads
        started is `max(n_consumer_threads,n_producer_threads)`.
    :type n_consumer_threads: int, optional
    :param consumer_group_id: the group ID under which each consumer should be created
    :type consumer_group_id: str, optional

    :raises ValueError: if `datafile_type` is not a subclass of
        :class:`~.data_file_io.DownloadDataFileToMemory`
    """

    def __init__(self,config_file,consumer_topic_name,producer_topic_name,**kwargs) :
        """
        Constructor method signature duplicated above to display in Sphinx docs
        """
        super().__init__(config_file,consumer_topic_name,producer_topic_name,**kwargs)

    def _get_processing_result_message_for_file(self,datafile,lock) :
        try :
            json_content = self._get_metadata_dict_for_file(datafile)
        except Exception as exc :
            errmsg = f'ERROR: failed to extract JSON metadata from {datafile.full_filepath}! '
            errmsg+= 'Error will be logged below, but not reraised.'
            self.logger.error(errmsg,exc_obj=exc,reraise=False)
            return None
        if not isinstance(json_content,dict) :
            errmsg = f'ERROR: JSON content found for {datafile.full_filepath} is of type {type(json_content)}, '
            errmsg+= 'not a dictionary! This file will be skipped.'
            self.logger.error(errmsg)
            return None
        return MetadataJSONMessage(datafile,json_content=json_content)

    @abstractmethod
    def _get_metadata_dict_for_file(self,datafile) :
        """
        Given a :class:`~.data_file_io.DownloadDataFileToMemory`, extract and return its metadata as a
        dictionary. The returned dictionary will be serialized to JSON and produced to the destination topic.

        This function can raise errors to be logged if metadata extraction doesn't proceed as expected.

        Not implemented in base class

        :param datafile: A :class:`~.data_file_io.DownloadDataFileToMemory` object that has received
            all of its messages from the topic
        :type datafile: :class:`~.data_file_io.DownloadDataFileToMemory`

        :return: metadata keys/values
        :rtype: dict
        """
        raise NotImplementedError

    def _on_shutdown(self) :
        self.logger.info('Will quit after all currently enqueued messages are received.')
        self.logger.info(self.progress_msg)
        super()._on_shutdown()

    @classmethod
    def run_from_command_line(cls,args=None) :
        """
        Run :func:`~produce_processing_results_for_files_as_read` to continually
        extract metadata from consumed files and produce them as json contents to another topic
        """
        # make the argument parser
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        xrd_csv_metadata_reproducer = cls(args.config,args.consumer_topic_name,args.producer_topic_name,
                                          consumer_group_id=args.consumer_group_id,
                                          n_consumer_threads=args.n_consumer_threads,
                                          n_producer_threads=args.n_producer_threads,
                                          output_dir=args.output_dir,
                                          update_secs=args.update_seconds,
                                          )
        # cls.bucket_name = args.bucket_name
        msg = f'Listening to the {args.consumer_topic_name} topic for XRD CSV files to send their metadata to the '
        msg+= f'{args.producer_topic_name} topic....'
        xrd_csv_metadata_reproducer.logger.info(msg)
        n_r,n_p,f_r_fns,m_p_fns = xrd_csv_metadata_reproducer.produce_processing_results_for_files_as_read()
        xrd_csv_metadata_reproducer.close()
        msg = f'{n_r} total message{"s were" if n_r!=1 else " was"} consumed'
        msg+=f', {n_p} message{"s were" if n_p!=1 else " was"} successfully processed'
        msg+=f', {len(f_r_fns)} file{"s were" if len(f_r_fns)!=1 else " was"} fully-read'
        if len(m_p_fns)>0 :
            msg+=f', and the following {len(m_p_fns)} file'
            msg+=' had' if len(m_p_fns)==1 else 's had'
            msg+=f' json metadata produced to the {args.producer_topic_name} topic:'
            for fn in m_p_fns :
                msg+=f'\n\t{fn}'
        xrd_csv_metadata_reproducer.logger.info(msg)



"""OpenMSIStream library for simplified data streaming using Apache Kafka"""

#imports
import os
from .data_file_io.entity.upload_data_file import UploadDataFile
from .data_file_io.actor.data_file_upload_directory import DataFileUploadDirectory
from .data_file_io.actor.data_file_download_directory import DataFileDownloadDirectory
from .data_file_io.actor.data_file_stream_processor import DataFileStreamProcessor
from .data_file_io.actor.data_file_stream_reproducer import DataFileStreamReproducer
from .s3_buckets.s3_transfer_stream_processor import S3TransferStreamProcessor

__all__ = [
    'UploadDataFile',
    'DataFileUploadDirectory',
    'DataFileDownloadDirectory',
    'DataFileStreamProcessor',
    'DataFileStreamReproducer',
    'S3TransferStreamProcessor',
]

#If running on Windows, try to pre-load the librdkafka.dll file
if os.name=='nt' :
    try :
        import confluent_kafka
    except Exception :
        import traceback
        reason = None
        import sys, pathlib
        dll_dir = pathlib.Path(sys.executable).parent/'Lib'/'site-packages'/'confluent_kafka.libs'
        if not dll_dir.is_dir() :
            reason = f'expected DLL directory {dll_dir} does not exist!'
        from ctypes import CDLL
        fps = []
        for fp in dll_dir.glob('*.dll') :
            if fp.name.startswith('librdkafka') :
                fps.append(fp)
                CDLL(str(fp))
        if len(fps)<1 :
            reason = f'{dll_dir} does not contain any librdkafka DLL files to preload!'
        if reason is not None :
            print(f'WARNING: Failed to preload librdkafka DLLs. Reason: {reason}')
        try :
            import confluent_kafka
        except Exception as e :
            errmsg = 'ERROR: Preloading librdkafka DLLs ('
            for fp in fps :
                errmsg+=f'{fp}, '
            errmsg = f'{errmsg[:-2]}) did not allow confluent_kafka to be imported! Exception (will be re-raised): '
            errmsg+= f'{traceback.format_exc()}'
            print(errmsg)
            raise e
    _ = confluent_kafka.Producer #appease pyflakes
